/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/*
 * cmdline.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>

#include <Performer/pf.h>
#include "pfutil.h"
#include "pfflt.h"
#include "pfsgi.h"

#include "perfly.h"

/* **************************************************************************
 *		Process the command line options
 ****************************************************************************/

static void
usage(char *progName)
{
    char	*baseName;
    static char *helpList =
	"usage:\n"
	"    %s [options] model.ext ...\n"
	"\n"
	"options:\n"
	"    -c <numChannels>              -Number of channels\n"
	"    -C a,b,c,...                  -Pipes for channels (R-to-L)\n"
	"    -d                            -Drive\n"
	"    -D                            -Lock down draw process\n"
	"    -e <h,p,r>                    -Initial view angles\n"
	"    -E                            -Earth sky off\n"
	"    -f                            -Fly\n"
	"    -F <path>                     -Set file path\n"
	"    -g <format>                   -GUI format\n"
	"        1 -> GUI_VERTICAL\n"
	"        2 -> GUI_HORIZONTAL\n"
	"	 0 -> Gui OFF\n"
	"    -G                            -Use GANGDRAW with Multipipe\n"
	"    -h                            -Print command list\n"
	"    -l <enable>		   -Set lighting enable\n"
	"	0 -> lighting off\n"
	"	1 -> lighting on\n"
	"    -l <x,y,z,r,g,b>              -Add a light source to the scene\n"
	"    -L                            -Free then Lock down processes (requires root ID)\n"
	"    -m <mode>                     -Multiprocessing mode\n"
	"       -1 -> PFMP_DEFAULT\n"
	"        0 -> PFMP_APPCULLDRAW\n"
	"        2 -> PFMP_APP_CULLDRAW\n"
	"        4 -> PFMP_APPCULL_DRAW\n"
	"        6 -> PFMP_APP_CULL_DRAW\n"
	"    -M <mode>                     -Multipipe mode\n"
	"        0 -> single pipe mode\n"
	"        1 -> multipipe mode\n"
	"        2 -> hyperpipe mode\n"
	"    -n <notify>                   -Debug level\n";
    static char *helpList2 =
	"    -N 			   -Assign non-degrading priorities to Performer processes\n"
	"    -o <mode>                     -Disable specified LoadFlt() mode\n"
	"    -O <mode>                     -Enable specified LoadFlt() mode\n"
	"    -p <x y z>                    -Initial view position\n"
	"    -P <phase>                    -Phase\n"
	"        0 -> PFPHASE_FLOAT\n"
	"        1 -> PFPHASE_LOCK\n"
	"        2 -> PFPHASE_FREE_RUN\n"
	"        3 -> PFPHASE_LIMIT\n"
	"    -r <frameRate>                -Target frame rate in Hertz\n"
	"    -R <mode>                     -Retessellate input geometry\n"
	"        0 -> Off\n"
	"        1 -> On\n"
	"    -s <LODscale>		   -Set LOD scale factor\n"
	"    -S <mode>                     -Show triangle strips via color coding\n"
	"        0 -> Off\n"
	"        1 -> On\n"
	"    -t                            -Trackball interface\n"
	"    -T                            -Earth sky uses TAG clear\n"
	"    -w                            -Write scene in ASCII\n"
	"    -W <xSize,ySize>              -Window size\n"
	"    -x                            -Toggle use of GLX with asynch X input handling\n"
	"    -X <radius>                   -Explode input files\n"
	"    -Z                            -Don't initially free CPUs before lock\n"
	"\n";

    /* just use program's base name: /usr/fred/base -> base */
    if (progName == NULL)
	baseName = "UNKNOWN";
    else
    if ((baseName = strchr(progName, '/')) == NULL)
	baseName = progName;

    /* print list of command-line options and arguments */
    fprintf(stderr, helpList, baseName);
    fprintf(stderr, helpList2);
    fflush(stderr);
}

char optionStr[] = "c:C:dDe:EfF:g:Ghl:Lm:M:o:O:Nn:p:P:r:R:s:S:TtW:wxX:Z?";

long
processCmdLine(int argc, char *argv[], char ***files)
{
    extern char *optarg;
    extern int	optind;
    long	opt;
    long	guiFormatSet = 0;

    /* process command-line arguments */
    while ((opt = getopt(argc, argv, optionStr)) != -1)
    {
	switch (opt)
	{
	/* Specify number of channels to use */
	case 'c':	
	    if (atoi(optarg) < 1)
		break;
	    NumChans = atoi(optarg);
	    if (!guiFormatSet && (NumChans > 1))
		ViewState->guiFormat = GUI_HORIZONTAL;
	    break;
	
	/* Specify the pipe to use for each channel */
	case 'C':	
	{
	    long	 index = 0;
	    long	 count = 0;
	    char	*start = optarg;

	    while (sscanf(start, "%d%n", &ChanOrder[index++], &count) == 1)
		if (*(start += count) != '\0')
		    ++start;
	}
	    break;
	
	/* Set pfuXformer default to Drive */
	case 'd':
	    ViewState->xformerMode = PFUXF_DRIVE;
	    break;
	
	/* lock down processor for just draw process */
	case 'D':
	    ViewState->procLock |= PFMP_FORK_DRAW;
	    break;
	    
	/* Set the initial view Euler angles (h p r) */
	case 'e':
	    if (sscanf(optarg, "%f,%f,%f",
		    &ViewState->initView.hpr[PF_H],
		    &ViewState->initView.hpr[PF_P],
		    &ViewState->initView.hpr[PF_R]) == 3)
		InitHPR = TRUE;
	    break;

	/* Turn off Earth Sky */
	case 'E':
	    ViewState->earthSkyMode = PFES_FAST;
	    break;

	/* Set pfuXformer default to Fly */
	case 'f':
	    ViewState->xformerMode = PFUXF_FLY;
	    break;

	/* Specify the file search path */
	case 'F':
	    pfFilePath(optarg);
	    break;

	/* Specify the GUI format */
	case 'g':
	    ViewState->gui = ((ViewState->guiFormat = atoi(optarg)) ? 1 : 0);
	    if( ViewState->guiFormat != GUI_HORIZONTAL )
		ViewState->guiFormat = GUI_VERTICAL;
	    guiFormatSet = 1;
	    break;
	   
	/* Initialize with GUI Panel disabled */
	case 'G':
	    GangDraw ^= 1;
	    break;
	
	/* remember light source definition */
	case 'l':
	{
	    float lxyz[3], lrgb[3], l;
	    if (sscanf(optarg, "%f,%f,%f,%f,%f,%f", 
		&lxyz[0], &lxyz[1], &lxyz[2], &lrgb[0], &lrgb[1], &lrgb[2]) == 6)
	    {
		if (ViewState->lamps < (MAX_LAMPS - 1))
		{
		    /* set default default values */
		    pfSetVec3(ViewState->lampXyz[ViewState->lamps], 0.0f,0.0f,0.0f);
		    pfSetVec3(ViewState->lampRgb[ViewState->lamps], 1.0f,1.0f,1.0f);
    
		    /* read specific values */
		    sscanf(optarg, "%f,%f,%f,%f,%f,%f",
			&ViewState->lampXyz[ViewState->lamps][0],
			&ViewState->lampXyz[ViewState->lamps][1],
			&ViewState->lampXyz[ViewState->lamps][2],
			&ViewState->lampRgb[ViewState->lamps][0],
			&ViewState->lampRgb[ViewState->lamps][1], 
			&ViewState->lampRgb[ViewState->lamps][2]);
    
		    /* advance lamp-definition count */
		    ViewState->lamps += 1;
		}
	    }
	    else
	    {
		if (sscanf(optarg, "%f", &l) == 1)
		    ViewState->lighting = l;
	    }
	}
	break;

	/* lock down processors for app, cull, and draw */
	case 'L':
	    ViewState->procLock = ((ulong)~0);
	    break;
	
	/* Specify multiprocessing mode */
	case 'm':
	    ProcSplit = atoi(optarg);
	    break;

	/* Enable multipipe mode */
	case 'M':
	    Multipipe = atoi(optarg);
	    break;
	    
	/* Set the notification (debug) level */
	case 'n':
	    NotifyLevel = atoi(optarg);
	    break;
	
	/* Assign non-degrading priorities to Performer processes. */
	case 'N':
	    PrioritizeProcs = 1;
	    break;
	
	case 'o':
	    LoadFltMode(atoi(optarg), 0);
	    break;
	
	case 'O':
	    LoadFltMode(atoi(optarg), 1);
	    break;

	/* Set the initial view position (x y z)*/
	case 'p':
	    if (sscanf(optarg, "%f,%f,%f",
		    &ViewState->initView.xyz[PF_X],
		    &ViewState->initView.xyz[PF_Y],
		    &ViewState->initView.xyz[PF_Z]) == 3)
		InitXYZ = TRUE;
	    break;

	/* Set the frame phase */
	case 'P':
	    ViewState->phase = atoi(optarg);
	    break;

	/* Set the frame rate */
	case 'r':
	    ViewState->frameRate = (float)atof(optarg);
	    if( ViewState->frameRate <= 0.0000001f)
		ViewState->frameRate = 30.0f;
	    break;

	/* Enable retessellation of input geometry */
	case 'R':
	    pfuMesherMode(PFUMESH_RETESSELLATE, atoi(optarg));
	    break;

	/* set LOD scale factor */
	case 's':
	    ViewState->LODscale = atof(optarg);
	    break;

	/* Enable color-coding of individual triangle strips */
	case 'S':
	    pfuMesherMode(PFUMESH_SHOW_TSTRIPS, atoi(optarg));
	    break;

	/* Use the RE TAG-Clear for Earth-sky */
	case 'T':
	    if (ViewState->realityGraphics)
		ViewState->earthSkyMode = PFES_TAG;
	    else
		ViewState->earthSkyMode = PFES_FAST;
	    break;

	/* Write scene to a file */
	case 'w':
	    WriteScene = TRUE;
	    break;

	/* Set the window size */
	case 'W':
	    if (sscanf(optarg, "%ld,%ld", &WinSizeX, &WinSizeY) != 2)
		if (sscanf(optarg, "%ld", &WinSizeX) == 1)
		    WinSizeY = WinSizeX;
	    break;

	case 'x':
	    ViewState->input = PFUINPUT_X;
	    break;

	/* explode files */
	case 'X':
	    ViewState->explode = (float)atof(optarg);
	    break;
	
	case 'Z':
	    FreeInitCPUs = 0;
	    break;

	/* Unknown option is no option at all */
	case '?':
	case 'h':
	default:
	    usage(argv[0]);
	    exit(1);
	}
    }

    /* return address of database filename list through "files" argument */
    *files = &(argv[optind]);

    /* function return value is number of database files to be read */
    return argc - optind;
}
