;;; driving.scm - Basic VR field with one or two cars.

(require 'persia)
(ps:dlopen "update_vehicle.so")

;;; SET SOME DEFAULT VALUES

;;; Should we initialize the simulator for using the VR helmet?
;;; Creates two windows in lower part of screen
(if (not (bound? 'doing-vr)) (define doing-vr #t))

;;; These next three flags allow one to selectively turn on/off sub-parts 
;;; of `doing-vr'.

;;; Open communication with the Polhemus or Flock and use it for camera pos'n.
(if (not (bound? 'doing-vr-head)) (define doing-vr-head doing-vr))
;;; Open communication with the Iscan Dell PC and get eye data from it
(if (not (bound? 'doing-vr-eyes)) (define doing-vr-eyes #f))
;;; Open communication with the kart, and get steering/gas/brake from it.
(if (not (bound? 'doing-kart)) (define doing-kart doing-vr))

;;; Create a second (mouse-controlled) car with a single console window
(if (not (bound? 'doing-second-car)) (define doing-second-car #f))

;;; Create a virtual camera that follows the first car from above
(if (not (bound? 'doing-vulture)) (define doing-vulture #f))

(if (not (bound? 'ground-width)) (define ground-width 2000.))
(if (not (bound? 'ground-depth)) (define ground-depth 2000.))
(if (not (bound? 'ground-z)) (define ground-z -5.))
(if (not (bound? 'first-car-x)) (define first-car-x 200.))
(if (not (bound? 'first-car-y)) (define first-car-y 200.))
(if (not (bound? 'second-car-x)) (define second-car-x 240.))
(if (not (bound? 'second-car-y)) (define second-car-y 200.))
(if (not (bound? 'first-car-control)) (define first-car-control 'accelerator))
(if (not (bound? 'second-car-control)) (define second-car-control 'mouse))
(if (not (bound? 'first-car-color)) (define first-car-color '#(.3 .1 .1)))
(if (not (bound? 'second-car-color)) (define second-car-color '#(.2 .2 .7)))
(if (not (bound? 'use-default-field)) (define use-default-field #t))
(if (not (bound? 'car-width)) (define car-width 2.5))
(if (not (bound? 'car-depth)) (define car-depth 5.))
(if (not (bound? 'car-height)) (define car-height 1.))
(if (not (bound? 'vulture-window-position))
    (define vulture-window-position '(50 990 650 500)))
(if (not (bound? 'second-car-window-position))
    (define second-car-window-position '(50 980 600 500)))


;;; LOAD HEAD-/HAND-POSITION AND KART INTERFACES
;;; (this must come before camera/window creation, because 
;;;  ps:position-windows-for-vr-eyes is defined in headhand.scm)

(if doing-vr
    (begin
      (if doing-kart
	  (require 'kart))
      (require 'headhand)
      (if doing-vr-eyes
	  (require 'iscan))
      ))

;;; CALCULATE NUMBER OF CAMERAS

(define num-cameras 0)

;;; This must come before ps:init-cameras
(if doing-vr
    (begin
      (ps:position-windows-for-vr-eyes 0 1)
      (define right-eye-camera 0)
      (define left-eye-camera 1)
      (set! num-cameras 2)))
(if doing-second-car
    (begin
      (apply ps:position-window (cons num-cameras second-car-window-position))
      (define second-car-camera num-cameras)
      (set! num-cameras (1+ num-cameras))))
(if doing-vulture
    (begin
      (apply ps:position-window (cons num-cameras vulture-window-position))
      (define vulture-camera num-cameras)
      (set! num-cameras (1+ num-cameras))))

(if (< num-cameras 1)
    (error 'num-cameras "doing-* flags must request at least one camera"))

;;; START GRAPHICS

(ps:init-cameras num-cameras)

;;; START READING DATA FROM THE HEADHAND AND KART
;;; (this must come after (ps:init-cameras))

(if doing-vr 
    (begin
      (if doing-vr-head
	  (ps:headhand-start))
      (if doing-vr-eyes
	  (ps:iscan-start))))

;;; SETUP TIME-GETTING

(ps:add-pre-frame-hook ps:set-delta-seconds-hook)

;;; MAKE WALLED FIELD

(if use-default-field
    (define field 
      (ps:make-dcs-walled-field
       ground-width ground-depth ground-z 
       (cons 'gstate (ps:make-gstate '(color . #(1. 1. 1. .4))
				     '(texture-file . "grass2.rgb")))
       (cons 'texture-scale 8.))))

;;; CREATE CARS

(define first-car-dcs 
  (if #f
      (get-esprit-dcs)
      (ps:make-dcs-car car-width car-depth car-height 
		       (cons 'gstate
			     (ps:make-gstate `(color . ,first-car-color))))))

(pf:dcs-trans first-car-dcs first-car-x first-car-y 0.)
(pf:add-child ps:*default-parent* first-car-dcs)

;;; Make a vehicle vector for the first car.
(define first-car-vehicle 
  (if doing-vr
      (begin
	;; Position the seat within the car.
	;; The origin of the car is the left-right-center, 
	;;  forward-back-center, on the floor.
	(define driver-headrest (pf:new-dcs))
	(pf:add-child first-car-dcs driver-headrest)
	(pf:dcs-trans driver-headrest 0. 0. (+ car-height .5))

	;; Create the head, and place it wrt to the headrest
	(if (bound? 'ps:*head-dcs*)
	    (begin
	      ;; We using a headhand position sensor
	      (define driver-head 
		(ps:place-vr-eyes-in-head-dcs
		 right-eye-camera left-eye-camera 0.025))
	      ;; We must insert the transformation from transmitter
	      ;; coordinates to headrest coordinates.
	      (let ((headhand-transmitter-to-headrest-dcs (pf:new-dcs)))
		(pf:dcs-mat headhand-transmitter-to-headrest-dcs
			    ps:*headhand-transmitter-to-headrest-pfmatrix*)
		(pf:add-child driver-headrest
			      headhand-transmitter-to-headrest-dcs)
		(pf:add-child headhand-transmitter-to-headrest-dcs
			      driver-head)))
	    (begin
	      ;; We are not using a headhand position sensor, but we are
	      ;; creating windows positioned for the VR helmet anyway;
	      ;; the virtual camera will not move with the wearer's head.
	      (define driver-head
		(ps:place-vr-eyes-in-dcs
		 right-eye-camera left-eye-camera 0.025 driver-headrest))
	      (pf:add-child driver-headrest 
			    driver-head)))

	;; When the AtoD-kart interface isn't working, make this #t
	(if (not doing-kart)
	    (display "Not connecting kart in driving.scm because bad AtoD\n")
	    (ps:make-kart-car-with-dcs first-car-dcs))
	)
      (if (not doing-second-car)
	  (ps:make-mouse-car-with-dcs first-car-dcs))))

(if doing-second-car
    (begin
      (define second-car-dcs 
	(ps:make-dcs-car car-width car-depth car-height 
			 (cons 'gstate
			       (ps:make-gstate `(color . ,second-car-color)))))
      (pf:dcs-trans second-car-dcs second-car-x second-car-y 0.)
      (pf:add-child ps:*default-parent* second-car-dcs)
      (define second-driver-headrest (pf:new-dcs))
;;;      (pf:dcs-trans second-driver-headrest -.5 -1. (+ car-height .5))
      (pf:dcs-trans second-driver-headrest .0 0. (+ car-height .35))
      (pf:add-child second-car-dcs second-driver-headrest)
      (define second-car-vehicle
	(ps:make-mouse-car-with-dcs second-car-dcs
				    (if (equal? second-car-control 'mouse)
					'mouse
					second-car-accel)))
      (ps:attach-camera second-car-camera second-driver-headrest)))

(if doing-vulture
    (begin
      (if doing-second-car
	  (ps:make-vulture-camera vulture-camera second-car-dcs 200.)
	  (ps:make-vulture-camera vulture-camera first-car-dcs 200.))
      (pf:chan-fov vulture-camera -1. 45.)))

(ps:set-do-pre-frame 1)

(provide 'driving)
