/* scm_engine_sound.c

   Engine noise is invoked with EngineSound(float pitch) where pitch can
   range from 0 (a very slow engine sound) to 1 (the fastest engine sound).
   If pitch is < 0 then the call to EngineSound() kills the engine
   noise.

   This works in conjunction with Jeff Pelz's miniboard code found on
   the Mac Dandelion, in the file
   Temporary Move Back to A40:Pelz:MiniBoard:

*/

#include "persia-scm.h"

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <stdlib.h>

#define DEVNAME "/dev/ttyd2"
#define BAUDRATE 9600
#define READTIMEOUT 10

struct SoundDriver {
  int speaker;
  int soundOn;

  char *deviceName;
  int baudRate;
  int readTimeout;
};

void
InitSoundDriver(struct SoundDriver *csd, 
		char *deviceName, 
		int baudRate, int readTimeout)
{
  struct termio term;
  int i;

  csd->deviceName = deviceName;
  csd->baudRate = baudRate;
  csd->readTimeout = readTimeout;
	
  if ((csd->speaker = open(csd->deviceName, O_WRONLY )) < 0) {
    perror(deviceName);
  }

  /* reset io flags */
  term.c_iflag = 0;
	
  /* no output post processing */
  term.c_oflag = 0;

  /* control flags - 8 bits processing, 19200 baud rate, enable receiver */
  if (csd->baudRate == 9600) {
    term.c_cflag = (B9600 | CS8 | CREAD | CLOCAL);
  }
  else {
    term.c_cflag = (B19200 | CS8 | CREAD | CLOCAL);
  }

  /* no local processing */
  term.c_lflag = 0;
  /* line handling ? */
/*  term.c_line = (unsigned char)0;*/
  term.c_line = LDISC1;
  
  /* no special chars */ 
  for(i = 0; i < NCCS; i++) {
    term.c_cc[i] = (unsigned char) 0;
  }

  /* set character read timeout, minimum characters to read is 1 */
  term.c_cc[VTIME] = (unsigned char) csd->readTimeout;
  term.c_cc[VMIN] = 0;

  if ((ioctl(csd->speaker, TCSETA, &term)) < 0) {
    perror("SoundDriver: Error in ioctl set dev char.\n");
  }

  csd->soundOn = 1;
}

int EngineSound (float arg)
{
  static struct SoundDriver csd;
  static int sound_driver_initialized = 0;

  int n_chars;
  char msg;

  if (!sound_driver_initialized) {
    InitSoundDriver(&csd, DEVNAME, BAUDRATE, READTIMEOUT);
    sound_driver_initialized = 1;
  }

  if (arg < 0)
    csd.soundOn = 0;
  else
    csd.soundOn = 1;

  if (csd.soundOn) {
    /* create message */
    msg = (unsigned char) (arg * 150);

    /* Send message */
    if((n_chars = write(csd.speaker, &msg, 1)) <= 0){
      perror("SoundDriver: Error in writing to device\n");
      return 0;
    }
  }
  else {
    /* Send stop message */
    /* Comment in Jeff's sound code says sending a 255 turns sound off. */
    msg = -1;
    if((n_chars = write(csd.speaker, &msg, 1)) <= 0)
      perror("SoundDriver: Error in writing to device\n");
    return 0;
  }
  return 1;
}

#ifndef ENGINE_SOUND_TEST

Object P_EngineSound (Object rpm)
{
  Check_Type(rpm, T_Flonum);
  EngineSound(FLONUM(rpm)->val);
  return Void;
}

void engine_sound_init ()
{
  Define_Primitive(P_EngineSound, "ps:engine-sound", 1, 1, EVAL);
}

#else

main ()
{
  const int MAXKEYRDSZ = 50;

  char buffer[MAXKEYRDSZ];

  EngineSound(4);
  printf("Press <return> to make it stop -> ");
  gets(buffer);
  EngineSound(-1);
}

#endif /* ENGINE_SOUND_TEST */
