/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/* 
 * env.c	environment functions, ie. weather, time of day, etc
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/18 17:41:21 $
 *
 */

#include <math.h>
#include <Performer/pf.h>
#include "pfutil.h"

#include "custom.h"

static pfFog	 *fog;
static pfVec3	fogColor;

#define	TOTAL_HORIZ_ANGLE	7.0f

	/*-----------------------------------------------------*/

void
initEnvironment(void)
{
    pfEarthSky 	*es;

    /* Initialize sun */
    ViewState->sun = pfNewLSource();

    pfAddChild(ViewState->scene, ViewState->sun);

    es = pfNewESky();
    pfESkyMode(es, PFES_BUFFER_CLEAR, ViewState->earthSkyMode); 
    pfESkyAttr(es, PFES_GRND_HT, -0.5f * ViewState->sceneSize);
    pfESkyColor(es, PFES_CLEAR, 0.5f, 0.5f, 0.8f, 1.0f);
    pfESkyAttr(es, PFES_HORIZ_ANGLE, TOTAL_HORIZ_ANGLE);

    /* setup the fog for general visibility */
    fog = pfNewFog(pfGetSharedArena());
    pfFogType(fog, ViewState->fog);
    pfFogColor(fog, 1.0f, 1.0f, 1.0f);
    pfSetVec3(fogColor, 1.0f, 1.0f, 1.0f);
    pfFogRange(fog, ViewState->nearFogRange, ViewState->farFogRange);
    
    ViewState->eSky = es;
}

void
updateEnvironment(void)
{
    pfEarthSky 		*es = ViewState->eSky;
    static long 	 eSkyMode = -1;
    static long 	 fogType = -1;
    static float 	 fogNear = 0.0f, fogFar = 0.0f;
    
    if ((eSkyMode == ViewState->earthSkyMode) && 
	(fogType == ViewState->fog) && 
	(fogNear == ViewState->nearFogRange) &&
	(fogFar == ViewState->farFogRange))
	return;
 
    if (ViewState->fog != PFFOG_OFF)
    {
	pfFogType(fog, ViewState->fog);
	pfFogRange(fog, ViewState->nearFogRange, ViewState->farFogRange);
	pfESkyFog(es, PFES_GENERAL, fog);
    }
    else
	pfESkyFog(es, PFES_GENERAL, NULL);

    pfESkyMode(es, PFES_BUFFER_CLEAR, ViewState->earthSkyMode); 
    
    /* update static values */
    fogType = ViewState->fog;
    fogFar = ViewState->farFogRange;
    eSkyMode = ViewState->earthSkyMode;
}


void
updateTimeOfDay(float newTod)
{
    float 		r, g, b;
    float 		s;
    pfVec3 		t;
    pfEarthSky		*es;
    pfLightSource	*sun;
    static float  	tod = -1.0f;
    static pfVec3	noon = {-.05f, -.1f, 1.0f};
    static pfVec3	night = {-.15f, 1.0f, .2f};

    /* only do if necessary */
    if(newTod == tod)
        return;

    tod = newTod;
    es = ViewState->eSky;
    sun = ViewState->sun;

    /* make horizon band angle directly proportional to time of day */
    pfESkyAttr(es, PFES_HORIZ_ANGLE, 1.0f + (TOTAL_HORIZ_ANGLE - 1.0f)*tod);

    /* scale it for the following ramp calcs */
    s = tod * 4.0f;	
    
    /* top of sky */
    r = (0.4f * s) - 1.6f;   r = PF_MAX2(0.0f, r);  r = PF_MIN2(0.7f, r);
    g = (0.4f * s) - 1.6f;   g = PF_MAX2(0.0f, g);  g = PF_MIN2(0.7f, g); 
    b = (0.8f * s) - 1.6f;   b = PF_MAX2(0.0f, b);  b = PF_MIN2(1.0f, b);
    pfESkyColor(es, PFES_SKY_TOP, r,g,b, 1.0f);
    
    /* bottom of sky */
    r = (0.6f * s) - 1.6f;   r = PF_MAX2(0.0f, r);  r = PF_MIN2(0.7f, r);
    g = (0.6f * s) - 1.6f;   g = PF_MAX2(0.0f, g);  g = PF_MIN2(0.7f, g); 
    b = (1.0f * s) - 1.6f;   b = PF_MAX2(0.0f, b);  b = PF_MIN2(1.0f, b);
    pfESkyColor(es, PFES_SKY_BOT, r,g,b, 1.0f);
	
    /* horizon */
    r = (1.45f * s) - 0.4f;   r = PF_MAX2(0.0f, r);  r = PF_MIN2(0.7f, r);
    g = (0.65f * s) - 1.4f;   g = PF_MAX2(0.0f, g);  g = PF_MIN2(0.7f, g); 
    b = (0.65f * s) - 1.6f;   b = PF_MAX2(0.0f, b);  b = PF_MIN2(1.0f, b);
    pfESkyColor(es, PFES_HORIZ, r,g,b, 1.0f);
    
    /* ground */
    r = (0.0181f * (s - 0.8f));   r = PF_MAX2(0.0f, r);  r = PF_MIN2(0.058f, r);
    g = (0.0840f * (s - 0.8f));   g = PF_MAX2(0.0f, g);  g = PF_MIN2(0.270f, g); 
    b = (0.0181f * (s - 0.8f));   b = PF_MAX2(0.0f, b);  b = PF_MIN2(0.058f, b); 
    pfESkyColor(es, PFES_GRND_NEAR, r ,g, b, 1.0f);
    pfESkyColor(es, PFES_GRND_FAR,  r ,g, b, 1.0f);

    /* Scale fog color based on time of day */
    if (ViewState->fog != PFFOG_OFF)
    {
	PFSCALE_VEC3(t, tod, fogColor);
	pfFogColor(fog, t[0], t[1], t[2]);
    }
    
    /* Here we set the light position. Alternately we could use a pfDCS. */
    pfCombineVec3(t, tod, noon, 1.0f - tod, night);
    pfNormalizeVec3(t);
    pfLightPos(sun, t[0], t[1], t[2], 0.0f);	
    pfLightAmbient(sun, 0.8f*tod, 0.8f*tod, 0.8f*tod); 
    pfLightColor(sun, 1.0f*tod, 1.0f*tod, 1.0f*tod); 
}
