/*
  Flock of Birds communications code.
  
  Assumptions:
  The host computer (running this code) connects to the master flock via RS232.
  
  Each flock (master and slaves) has a unique FBB address.  The host can
  communicate to the slave flock boxes via the FBB to the master.  The
  host prefixes flock commands with the RS232 To FBB command to accomplish
  this.
  
  Read "The Flock of Birds Installation and Operation Guide" for info
  on how the box works (and therefore some of why this code works the
  way it does).
  */

/* Page number refer to 
   "The Flock of Birds(TM) Installation and Operation Guide" 
   July 5, 1993 */

#include "persia-scm.h"
#include "persia.h"
#include "flock.h"

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
/* #include "LinearAlgebra.h" */
#include <assert.h>
#ifndef sgiport
#include "/u/becker/src/sgi_vr/Ptvme340/App/hostcmd_sgi.h"
#endif

#ifdef sgiport
/*#define FLOCK_DEVICE_NAME	"/dev/ttyd2"*/
#else
#define FLOCK_DEVICE_NAME	"0"
#endif
#define READ_TIMEOUT		10
#define SUCCESS 0 
#define FAIL    (-1) 
#define FLOCK 1			/* used as flag to the read_n_bytes routine */
#define HEMI_FORWARD	0
#define HEMI_AFT	1
#define HEMI_UPPER	2
#define HEMI_LOWER	3
#define HEMI_LEFT	4
#define HEMI_RIGHT	5
#define REQUEST_DATA 'B'

#define POSK36 (float)(36.0/32768.0)
#define POSK72 (float)(72.0/32768.0)
#define POSK144 (float)(144.0/32768.0)
#define ANGK   (float)(180.0/32768.0)
#define MATK	(float)(0.99996/32768.0)

typedef char BOOL;
#define YES ((BOOL)1)
#define NO ((BOOL)0)

/* Yucky globals.  Used to be instance variables. */
static int fd;
static int positionScaling;
static BOOL directConnect;
static char rs232ToFBBCommand;

static int num_receivers = 1;	/* may be either 1 or 2 */

/* These values based on position-matrix mode */
#define ONE_FLOCK_RECORD_LENGTH (24)
#define TWO_FLOCK_RECORD_LENGTH (24 + 1 + 24 + 1)

static int flock_cnvrt_flock_to_int(char LSB, char MSB)
{
  short n;
  n = (short) (((short) (MSB << 8) | (short) ((LSB & 0x7f) << 1)) << 1);
  return((int) n);
}

void flock_initialize_devname(char *deviceName, int baudRate)
{
  directConnect = YES;
  
  if (flock_open_port(deviceName, baudRate) == FAIL) {
    fprintf(stderr, "flock_constructor's openPort() failed\n");
    exit(1);
  }
  
  /* XXX This isn\'t very generic -- think about changing it. */
  /* Change value cmd = \'P\'
     AUTO-CONFIG = 50
     number of units (master+slave) = 2 (or 3 if using extended
     range transmitter)
     */
#ifdef extended_range_transmitter
  flock_send_n_bytes("P\062\003", 3);
#else
  flock_send_n_bytes("P\062\002", 3);
#endif

#ifdef extended_range_transmitter
  /* Next transmitter cmd = \'060\'
     addr/number = \'060\'
   */
  send_n_bytes("\060\060", 2);
#endif
  
  flock_initialize();
  
  /* this should not go here -- call setFilters() outside here */
#ifdef hack
  /* turn filters off
     setFilters(7); */

  /* turn filters on
     setFilters(4); */
  flock_set_filters(7);
#endif
  /* eventID = FLOCK_EVENT; */
}

void flock_initialize_fbb(int FBBAddress, int fileDescriptor)
{
  directConnect = NO;
  rs232ToFBBCommand = 0xf0 + FBBAddress;
  fd = fileDescriptor;
  flock_initialize();
  /* eventID = FLOCK_SLAVE_EVENT; */
}

void flock_initialize()
{
#ifdef extended_range_transmitter
  flock_set_position_scaling(144);
#else
  flock_set_position_scaling(36);
#endif
  flock_set_position_matrix_mode();
  flock_set_bird_hemisphere(HEMI_LOWER);
  /* flock_set_num_receivers(1); */
}

int flock_num_receivers()
{
  return num_receivers;
}

void flock_set_num_receivers(int n)
{
  if (n < 1 || n > 2)
    {
      fprintf(stderr, "%s:%d: Bad number of receivers\n", __FILE__, __LINE__);
      exit(-1);
    }
  num_receivers = n;

  if (num_receivers == 2)
    {
      char buf[8];

      /* Put the flock in GROUP mode, see page 37 and page 52. */
      /* "P\x23\x1" */
      flock_send_n_bytes("P\043\001", 3);

#define CHECK_GROUP_MODE 0
#if CHECK_GROUP_MODE
      /* Make sure it was successful with an EXAMINE VALUE command */
      flock_send_n_bytes("O\043", 2);
      flock_read_n_bytes(buf, 1, 0);
      assert(*buf == 1);
#endif
    }
  else
    {
      char buf[8];

      /* Disable GROUP mode, see page 37 and page 52. */
      /* "P\x23\x0" */
      flock_send_n_bytes("P\043\000", 3);

#if CHECK_GROUP_MODE
      /* Make sure it was successful with an EXAMINE VALUE command */
      flock_send_n_bytes("O\043", 2);
      flock_read_n_bytes(buf, 1, 0);
      assert(*buf == 0);
#endif
    }
}

#if 0
flock_create_slave(int FBBAddress)
{
  FlockComm *slave;
  slave = flock_initialize_fbb(FBBAddress, fd);
  return (slave);
}
#endif

int flock_open_port(char *devName, int baudRate)
{
#ifdef sgiport
  struct termio term;
  int i;
  
  if ((fd = open(devName, O_RDWR )) < 0) {
    perror(devName);
    return(FAIL);
  }
  
  /* reset io flags */
  term.c_iflag = 0;
  
  /* no output post processing */
  term.c_oflag = 0;
  
  /* control flags - 8 bits processing, 19200 baud rate, enable receiver */
  term.c_cflag = (CS8 | CREAD | CLOCAL);
  switch(baudRate) {
  case 2400:
    term.c_cflag |= B2400;
    break;
  case 4800:
    term.c_cflag |= B4800;
    break;
  case 9600:
    term.c_cflag |= B9600;
    break;
  case 19200:
    term.c_cflag |= B19200;
    break;
  case 38400:
    term.c_cflag |= B38400;
    break;
  }

  /* no local processing */
  term.c_lflag = 0;
  /* line handling ? */
  term.c_line = (unsigned char)0;
	
  /* no special chars */ 
  for(i=0;i<NCCS;i++)term.c_cc[i]=(unsigned char)0;

  /* set character read timeout, minimum characters to read is 0 */
  term.c_cc[VTIME] = (unsigned char)READ_TIMEOUT;  
  term.c_cc[VMIN] = 0;
	
  if ((ioctl(fd, TCSETA, &term)) < 0) {
    fprintf(stderr, "flock_open_port: error in ioctl to set flock device characteristics\n");
    return(FAIL);
  }
#else
  hostcmd_init();
  iusc_init(atoi(devName), baudRate);

  fd = atoi(devName);
#endif

  return(SUCCESS);
}

void flock_get_matrix(float trans[3], float rot[3][3])
{
  char dataBuf[ONE_FLOCK_RECORD_LENGTH];
  int i, j;
  int index;

  if (directConnect == NO)
    flock_send_byte(rs232ToFBBCommand);

  flock_send_byte(REQUEST_DATA);
  assert (num_receivers == 1);
#ifdef sgiport
  flock_read_n_bytes(dataBuf, ONE_FLOCK_RECORD_LENGTH, FLOCK);
#else
  if (read_340iusc_last_pkt(fd, dataBuf, ONE_FLOCK_RECORD_LENGTH) == 0) {
    fprintf(stderr,"flock_get_matrix: Timed out on read\n");
    perror("flock_get_matrix");
    return;
  }
#endif

  switch (positionScaling) {
  case 36:
    trans[0] = (float)flock_cnvrt_flock_to_int(dataBuf[0],dataBuf[1]) * POSK36;
    trans[1] = (float)flock_cnvrt_flock_to_int(dataBuf[2],dataBuf[3]) * POSK36;
    trans[2] = (float)flock_cnvrt_flock_to_int(dataBuf[4],dataBuf[5]) * POSK36;
    break;
  case 72:
    trans[0] = (float)flock_cnvrt_flock_to_int(dataBuf[0],dataBuf[1]) * POSK72;
    trans[1] = (float)flock_cnvrt_flock_to_int(dataBuf[2],dataBuf[3]) * POSK72;
    trans[2] = (float)flock_cnvrt_flock_to_int(dataBuf[4],dataBuf[5]) * POSK72;
    break;
  case 144:
    trans[0] = (float)flock_cnvrt_flock_to_int(dataBuf[0],dataBuf[1])* POSK144;
    trans[1] = (float)flock_cnvrt_flock_to_int(dataBuf[2],dataBuf[3])* POSK144;
    trans[2] = (float)flock_cnvrt_flock_to_int(dataBuf[4],dataBuf[5])* POSK144;
    break;
  }

  /* matrix data starts at array index 5 */
  index = 5;

  for (i=0;i<3;++i)
    for (j=0;j<3;++j)
      rot[i][j] = (float)
	flock_cnvrt_flock_to_int(dataBuf[++index], dataBuf[++index]) * MATK;
}

void flock_get_two_matrices(float trans1[3], float rot1[3][3],
			    float trans2[3], float rot2[3][3])
{
  char dataBuf[TWO_FLOCK_RECORD_LENGTH];
  int i, j;
  int index;
  float scaling;

  if (directConnect == NO)
    flock_send_byte(rs232ToFBBCommand);

  flock_send_byte(REQUEST_DATA);
  assert (num_receivers == 2);
#if sgiport
  flock_read_n_bytes(dataBuf, TWO_FLOCK_RECORD_LENGTH, FLOCK);
#else
  if (read_340iusc_last_pkt(fd, dataBuf, TWO_FLOCK_RECORD_LENGTH) == 0) {
    fprintf(stderr,"flock_get_matrix: Timed out on read\n");
    perror("flock_get_matrix");
    return;
  }
#endif

#define PRINT_DATA 0
#if PRINT_DATA
  for (i = 0; i < 50; i++)
    printf("0x%x ", (unsigned) dataBuf[i]);
  printf("\n");
#endif

  switch (positionScaling) {
  case 36:
    scaling = POSK36;
    break;
  case 72:
    scaling = POSK72;
    break;
  case 144:
    scaling = POSK144;
    break;
  }

  index = 0;
  trans1[0] = 
    flock_cnvrt_flock_to_int(dataBuf[0],dataBuf[1]) * scaling;
  trans1[1] = 
    flock_cnvrt_flock_to_int(dataBuf[2],dataBuf[3]) * scaling;
  trans1[2] = 
    flock_cnvrt_flock_to_int(dataBuf[4],dataBuf[5]) * scaling;

  index = 6;
  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
	{
	  rot1[i][j] = (float)
	    flock_cnvrt_flock_to_int(dataBuf[index], dataBuf[index+1]) * MATK;
	  index += 2;
	}

  index++;			/* For group mode address byte, see page 52 */

  trans2[0] = 
    flock_cnvrt_flock_to_int(dataBuf[index],dataBuf[index+1]) * scaling;
  index += 2;
  trans2[1] = 
    flock_cnvrt_flock_to_int(dataBuf[index],dataBuf[index+1]) * scaling;
  index += 2;
  trans2[2] = 
    flock_cnvrt_flock_to_int(dataBuf[index],dataBuf[index+1]) * scaling;
  index += 2;

  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
	{
	  rot2[i][j] = (float)
	    flock_cnvrt_flock_to_int(dataBuf[index], dataBuf[index+1]) * MATK;
	  index += 2;
	}

#define PRINT_MATRIX 0
#if PRINT_MATRIX
  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      printf("%f ", rot1[i][j]);
  printf("\n");
  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      printf("%f ", rot2[i][j]);
  printf("\n");

#endif
}

/*
   Send a one-byte message to fdesc
*/
int flock_send_byte(char msg)
{
  int n_chars;

  /* Send message */
#ifdef sgiport
  if((n_chars=write(fd, &msg, 1))<=0){
#else
  if((n_chars=write_340iusc(fd, &msg, 1))<=0){
#endif
    fprintf(stderr,"flock_send_byte: Error in writing to device\n");
    perror("send_byte");
    return(FAIL);
  }
  return(SUCCESS);
}

/*
   Send a multi-byte message to fdesc
*/
int flock_send_n_bytes(char *msg, int n_bytes)
{
  int n_chars;

  /* Send message */
#ifdef sgiport
  if((n_chars=write(fd, msg, n_bytes)) != n_bytes){
#else
  if((n_chars=write_340iusc(fd, msg, n_bytes)) != n_bytes){
#endif
    fprintf(stderr,"flock_send_n_bytes: Error in writing to device\n");
#ifdef sgiport
    perror("flock_send_n_bytes");
#endif
    return(FAIL);
  }
  return(SUCCESS);
}

/*
   Read N bytes from the fdesc 
   the parameter check_flock_bit determines whether bit 7 is checked
   (to determine whether the flock is in sync)
*/
int flock_read_n_bytes(char *ch, int n_bytes, int check_flock_bit)
{
  int n_chars;
  int i;

  /* Read n_bytes from serial port */
  /* If less than n_bytes are read, keep reading until they are all read */
  n_chars = 0;
  while( n_chars < n_bytes ) {
    /* Was there an error while reading? */
#ifdef sgiport
    if( (i = read(fd,ch+n_chars,n_bytes-n_chars)) < 0 ) {
#else
    if( (i = read_340iusc(fd,ch+n_chars,n_bytes-n_chars)) < 0 ) {
#endif
      fprintf(stderr,"flock_read_n_bytes: Error reading serial port");
#ifdef sgiport
      perror("flock_read_n_bytes");
#endif
      return(FAIL);
    }
    /* Did read timeout? */
    if( i == 0 ) {
      fprintf(stderr,"flock_read_n_bytes: Timed out on read (got %d, expected %d bytes\n", n_chars, n_bytes);
#ifdef sgiport
      perror("flock_read_n_bytes");
#endif
      return(FAIL);
    }
    n_chars += i;
  }
  /* If in check_flock_bit mode, Check to see if in sync */
  if( (check_flock_bit == FLOCK) && !(ch[0] & 128) ) {
      fprintf(stderr,"flock_read_n_bytes: Out of Sync. Bit 7 on LSB is off.");
      perror("read_n_bytes");
      return(FAIL);
  }

  return(n_chars);
}

/*******************************************************************/
/*  When a single bird is used, the position scaling is set with the P3 cmd */
/*  For a flock of birds, preface with the RS232 to FOBB cmd (242) */
void flock_set_position_scaling (int posScaling)
{
  positionScaling = posScaling;

  if (directConnect == NO)
    flock_send_byte(rs232ToFBBCommand);

  /* Switch BIRDs position scale to 36 or 72 inches */
  flock_send_byte('P');
  flock_send_byte(3);
  if(positionScaling == 72 ) {
    flock_send_byte(1);
    flock_send_byte(0);
  }
  else {
    flock_send_byte(0);
    flock_send_byte(0);
  }
}

void flock_send_bird_hemisphere_values(int hemisphere)
{
  switch(hemisphere) {
  case HEMI_FORWARD:
    flock_send_byte(0);
    flock_send_byte(0);
    break;
  case HEMI_AFT:
    flock_send_byte(0);
    flock_send_byte(1);
    break;
  case HEMI_UPPER:
    flock_send_byte(0xc);
    flock_send_byte(1);
    break;
  case HEMI_LOWER:
    flock_send_byte(0xc);
    flock_send_byte(0);
    break;
  case HEMI_LEFT:
    flock_send_byte(6);
    flock_send_byte(1);
    break;
  case HEMI_RIGHT:
    flock_send_byte(6);
    flock_send_byte(0);
    break;
  }
}

void flock_set_bird_hemisphere(int hemisphere)
{
  if (directConnect == NO)
    flock_send_byte(rs232ToFBBCommand);

  flock_send_byte('L');
  flock_send_bird_hemisphere_values(hemisphere);

  flock_send_byte(0xf2);
  flock_send_byte('L');
  flock_send_bird_hemisphere_values(hemisphere);
}

/* put in position/matrix mode */
void flock_set_position_matrix_mode()
{
  if (directConnect == NO)
    flock_send_byte(rs232ToFBBCommand);

  flock_send_byte('Z');

  /* set up slave, see page 70 */
#if 1
  flock_send_n_bytes("\362Z", 2);
#endif
}

void flock_set_filters(char filterBits)
{
  if (directConnect == NO)
    flock_send_byte(rs232ToFBBCommand);

  flock_send_n_bytes("P\004", 2);
  flock_send_byte(filterBits);
  flock_send_byte(0);
}


/* Scheme Interface */

static unsigned elk_num_receivers = 1;

/* Some exported variables */
pfDCS *flock_helmet_dcs;
pfDCS *flock_hand_dcs;

static int doing_flock;
static int flock_initialized;

void flock_initialize_scm()
{
  static int done = 0;
  if (!done)
    {
      done = 1;
      flock_helmet_dcs = pfNewDCS();
      flock_hand_dcs = pfNewDCS();
      Define_Pfo((pfObject*)flock_helmet_dcs, "ps:*head-dcs*");
      Define_Pfo((pfObject*)flock_hand_dcs, "ps:*hand-dcs*");
    }
}

void flock_c_hook()
{
  /* The helmet coordinate system is:
     with helmet on head, X is to your right, Y is up, 
     and Z is to your rear.	*/
  /* fooInBar, means that we describe foo's axis in terms of
	   bar's coordinate systems, which means that in the first row
	   maps foo's x-axis to bar's.  If foo has x where bar has -y,
	   then the first row is 0 -1 0 */
      /* I define "WorldCoords" to have 
	   origin on the floor below the transmitter
	   +x towards Macintosh wall
	   +y towards the curved wall
	   +z up. */
  static pfMatrix flockXmitInWorldCoords = {
    { 0, 1, 0, 0},
    { 1, 0, 0, 0},
    { 0, 0, -1, 0},
    { 0, 0, 1.43, 1}};

  float trans[3];
  float rot[3][3];
  float trans2[3];
  float rot2[3][3];
  int i, j;
  pfMatrix flockRcvInFlockXmitCoords;
  int num_birds;

  assert (doing_flock);

  if (!flock_initialized)
    {
      flock_initialize_scm();
#if sgiport
      flock_initialize_devname("/dev/ttyd2", 19200);
#else
      flock_initialize_devname("0", 57600);
#endif
      flock_initialized = 1;
    }

  if ((num_birds = flock_num_receivers()) != elk_num_receivers)
    {
      flock_set_num_receivers(elk_num_receivers);
      num_birds = elk_num_receivers;
    }
  if (num_birds == 2)
    flock_get_two_matrices(trans, rot, trans2, rot2);
  else
    flock_get_matrix(trans, rot);

  flockRcvInFlockXmitCoords[3][0] = trans[0] * 0.0254;
  flockRcvInFlockXmitCoords[3][1] = trans[1] * 0.0254;
  flockRcvInFlockXmitCoords[3][2] = trans[2] * 0.0254;
  flockRcvInFlockXmitCoords[0][3] = 0.0;
  flockRcvInFlockXmitCoords[1][3] = 0.0;
  flockRcvInFlockXmitCoords[2][3] = 0.0;
  flockRcvInFlockXmitCoords[3][3] = 1.0;
  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      flockRcvInFlockXmitCoords[i][j] = rot[j][i];
  /* fprintf(stderr, "%f %f %f\n", trans[0], trans[1], trans[2]); */
  {
    pfMatrix flockRcvInWorldCoords;
    pfMultMat(flockRcvInWorldCoords,
	      flockRcvInFlockXmitCoords,
	      flockXmitInWorldCoords);
    pfDCSMat(flock_helmet_dcs, flockRcvInWorldCoords);
  }

  if (num_birds == 2)
    {
      flockRcvInFlockXmitCoords[3][0] = trans2[0] * 0.0254;
      flockRcvInFlockXmitCoords[3][1] = trans2[1] * 0.0254;
      flockRcvInFlockXmitCoords[3][2] = trans2[2] * 0.0254;
      flockRcvInFlockXmitCoords[0][3] = 0.0;
      flockRcvInFlockXmitCoords[1][3] = 0.0;
      flockRcvInFlockXmitCoords[2][3] = 0.0;
      flockRcvInFlockXmitCoords[3][3] = 1.0;
      for (i = 0; i < 3; i++)
	for (j = 0; j < 3; j++)
	  flockRcvInFlockXmitCoords[i][j] = rot2[j][i];
      /* fprintf(stderr, "%f %f %f\n", trans[0], trans[1], trans[2]); */
      {
	pfMatrix flockRcvInWorldCoords;
	pfMultMat(flockRcvInWorldCoords,
		  flockRcvInFlockXmitCoords,
		  flockXmitInWorldCoords);
	pfDCSMat(flock_hand_dcs, flockRcvInWorldCoords);
      }
    }
}

Object P_StartFlock()
{
  if (!ViewState)
    Primitive_Error("ps:init-cameras not yet called");
  if ( ! doing_flock )
    {
      flock_initialize_scm();
      doing_flock = 1;
      persia_add_pre_frame_hook(flock_c_hook);
    }
  return Void;
}

Object P_StopFlock()
{
  if (!ViewState)
    Primitive_Error("ps:init-cameras not yet called");
  if (doing_flock)
    {
      doing_flock = 0;
      persia_remove_pre_frame_hook(flock_c_hook);
    }
  return Void;
}

Object P_SetFlockNumReceivers(Object num)
{
  Check_Type(num, T_Fixnum);
  if (FIXNUM(num) < 1 || FIXNUM(num) > 2)
    Primitive_Error("number of flock recievers out of range");
  elk_num_receivers = FIXNUM(num);
  return Void;
}

void flock_init()
{
  Define_Primitive(P_StartFlock, "ps:headhand-start", 0, 0, EVAL);
  Define_Primitive(P_StopFlock, "ps:headhand-stop", 0, 0, EVAL);
  Define_Primitive(P_SetFlockNumReceivers, "ps:headhand-set-num-receivers", 
		   1, 1, EVAL);
}
