/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 *	generic.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <sys/sysmp.h>

/* X Window includes */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <gl/glws.h>

/* GL includes */
#include <gl/device.h>

/* IRIS Performer includes */
#include <Performer/pf.h>
#include "pfutil.h"
#include "pfflt.h"
#include "pfsgi.h"

/* common includes */
#include "main.h"
#include "custom.h"

#if MCCALLUM
#include "elk.h"
#include <assert.h>
#endif

/******************************************************************************
 *			    Global Variables
 *****************************************************************************/

/* Data shared between processes */
SharedViewState	*ViewState;

/*
 * Command line optional settings for initialization. These variables are
 * set before the call to pfConfig and are only used for initialization.
 * So they don't need to be in the ViewState shared memory structure. They
 * are initialized in the function InitSharedMem().
*/
long	WinSizeX		= -1;
long	WinSizeY		= -1;
long	Multipipe		= -1;
long	NumPipes		= -1;
long	NumChans		= -1;
long	NotifyLevel		= PFNFY_INFO;
#if 0 && MCCALLUM
/* Temporarily turn off multiprocessing. */
long	ProcSplit		= PFMP_APPCULL_DRAW;
#else
long	ProcSplit		= PFMP_DEFAULT;
#endif
long	PrioritizeProcs		= 0;
long	GangDraw		= 0;
long	InitHPR			= FALSE;
long	InitXYZ			= FALSE;
long 	WriteScene		= FALSE;
long	ChanOrder[MAX_CHANS]	= {-1};

/*********************************************************************
*								     *
*		APPLICATION PROCESS ROUTINES			     *
*								     *
**********************************************************************/

/* The following globals are used by the Application process only */

#if MCCALLUM
#define Chan ViewState->chans
#else
pfChannel   *Chan[MAX_CHANS];
#endif

long	    NumFiles	    = 0;
char	    **DatabaseFiles = NULL;

/* Determine if Simulation is finished */
long
SimDone(void)
{
    return(ViewState->exitFlag);
}

/*
 * Allocate Global pointers and set Global variables-
 * This must be called before pfConfig() if forked processes
 * must share globals and/or pointers to shared memory.
 * Initialize shared memory structures
*/
void
InitSharedMem(int argc, char *argv[], int num_cameras)
{
    int		c;
    long 	maxScreenX;
    long 	maxScreenY;

    /*
     * Malloc structures now, before pfConfig(), so forked processes have same
     * pointers.
     */
    /* Do this before start_graphics, so that elk has a place to
       put the window position preferences */
    ViewState = (SharedViewState *) pfCalloc(1, sizeof(SharedViewState), 
					     pfGetSharedArena());

    /* Find out how big the window can be */
    maxScreenX = getgdesc(GD_XPMAX);
    maxScreenY = getgdesc(GD_YPMAX);

    /*
     * Set window size - this is set here, before the call to pfConfig,
     * so that all processes will get the same copy. These do not need
     * to be in shared memory, because they are initialization values
     * which are not changed during the course of the application
    */
    if ((WinSizeX < 0) || (WinSizeY < 0))
    {	
	/* If no window size specified, set to be max possible */
	WinSizeX   = maxScreenX;
	WinSizeY   = maxScreenY;
    }
    else
    {	
	/* If the window size was set by command line arguments */
	WinSizeX = PF_MIN2(WinSizeX,maxScreenX);
	WinSizeY = PF_MIN2(WinSizeY,maxScreenY);
    }

    /* Initialize the channel-pipe assignment table: 0->0, 1->1, ... */
    for (c = 0; c < MAX_CHANS; c++)
	ChanOrder[c] = c;

    /* Initialize Keyboard and GUI ViewState data */
    initViewState(num_cameras);

#if MCCALLUM
    Multipipe = TRUE;
    NumPipes = ViewState->num_cameras;
#else
    /* Set default multipipe mode */
    Multipipe = (getgdesc(GD_NSCRNS) > 1);
#endif

    /* Process the command line args */
    NumFiles = processCmdLine(argc, argv, &DatabaseFiles);
    if (!ViewState->guiFormat)
	ViewState->guiFormat = GUI_VERTICAL;

    /* Figure out the number of channels and pipelines required */
    switch (Multipipe) {
    case TRUE:
#if MCCALLUM
        NumChans = NumPipes;
#else
	if (NumChans < 0)
	    NumChans = getgdesc(GD_NSCRNS);
	NumPipes = NumChans;
#endif
	if (!ViewState->guiFormat)
	    ViewState->guiFormat = GUI_HORIZONTAL;
	break;
    case FALSE:
	if (NumChans < 0)
	    NumChans = 1;
	else if ((NumChans > 1) && (!ViewState->guiFormat))
		ViewState->guiFormat = GUI_HORIZONTAL;
	NumPipes = 1;
	break;
    case HYPERPIPE:
	/* MUXPIPES not implemented for RealityEngine */
	if (0 && !getgdesc(GD_MUXPIPES))
	{
	    pfNotify(PFNFY_WARN, PFNFY_RESOURCE,
			"This machine not is not capable of Hyperpipe mode");

	    Multipipe = TRUE;
	}
	else
	{
	    NumPipes = getgdesc(GD_NSCRNS);
	    NumChans = 1;
	}
	if (!ViewState->guiFormat)
	    ViewState->guiFormat = GUI_HORIZONTAL;
	break;
    }
}



/* Initialize MP, system config */
void
InitConfig(void)
{
    long 	numCPUs = 0;
    long 	availCPUs;

    /* set a path to everywhere we want to look for data */
    if (pfGetFilePath() == NULL)
	pfFilePath(
		   "."
		   ":../data"
		   ":../../data"
		   ":../../data/polyhedra"
		   ":../../../data"
		   ":/usr/src/Performer/data"
		   ":/usr/demos/data/models"
		   ":/usr/demos/data/flt"
#if MCCALLUM
		   ":/usr/src/Performer/friends/models/iv"
		   ":/usr/src/Performer/friends/site"
		   ":/usr/src/Performer/friends/models/flt"
		   ":/usr/src/Performer/friends/models/bin"
		   ":/usr/src/Performer/friends/models/dwb"
		   ":/usr/src/Performer/friends/models/sv"
		   ":/usr/src/Performer/friends/paradigm/bfiles"
		   ":/usr/src/Performer/friends/town/toon/pftexture"
		   ":/usr/src/Performer/friends/town/toon/pfmodel"
		   ":/usr/src/Performer/friends/town/orig/pfmodel"
		   ":/usr/src/Performer/friends/town/orig/pftexture"
		   ":/usr/src/Performer/friends/viewpoint/1992"
		   ":/usr/src/Performer/friends/viewpoint/1993"
		   ":/usr/src/Performer/friends/avalon/misc"
		   ":/usr/src/Performer/friends/avalon/planes"
		   ":/usr/src/Performer/friends/avalon/satellites"
		   ":/usr/src/Performer/friends/avalon/space"
		   ":/usr/src/Performer/friends/avalon/terrain"
		   ":/usr/src/Performer/friends/avalon/vehicles"
		   ":/usr/src/Performer/friends/caryphaeus/MODELS/3D_MODELS"
#endif
		   );
    /* Set Debug Level */
    pfNotifyLevel(NotifyLevel);

    /* Set multiprocessing mode. */
    pfMultiprocess(ProcSplit);

    /* Set number of pfPipes desired. */
    pfMultipipe(NumPipes);

    /* Enable hyperpipe mode */
    if (Multipipe == HYPERPIPE)
    	pfHyperpipe(NumPipes);

    /* Set intersection callback. */
    pfIsectFunc(IsectFunc);

    /* Check for restricted CPUs */
    availCPUs = sysmp(MP_NAPROCS);
    numCPUs = sysmp(MP_NPROCS);

    if (numCPUs != availCPUs)
    {
	pfNotify(PFNFY_NOTICE, PFNFY_RESOURCE, 
	    "Only %d out of %d processors available on this machine !!!",
			availCPUs, numCPUs);

	if ((geteuid() == 0) && FreeInitCPUs)
	    pfuFreeCPUs();
	else
	    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "Not freeing CPUs.");
    } 
    else
	pfNotify(PFNFY_INFO, PFNFY_PRINT, 
	    "All %d processors available on this machine.", numCPUs);
}


/* Initialize the visual database */
void
InitScene(void)
{
    int		loaded =0;
    pfScene	*scene;

    /* Read in any databases specified on cmd line */
    ViewState->scene = scene = pfNewScene();

    loaded = initSceneGraph(scene);

    /* Set up scene graph for collisions */
#if MCCALLUM
    #if DOING_COLLISION_DETECTION
    pfuCollideSetup(scene, PFUCOLLIDE_DYNAMIC, 0xffffffff);
    #endif
#else
    pfuCollideSetup(scene, PFUCOLLIDE_STATIC, 0xffffffff);
#endif

    /* Initialize EarthSky, fog, and sun model */
    initEnvironment();

    /* Write out nodes in scene (for debugging) */
    if (WriteScene)
    {
	FILE *fp;
	if (fp = fopen("scene.out", "w"))
	{
	    pfPrint(scene, PFTRAV_SELF|PFTRAV_DESCEND, PFPRINT_VB_OFF, fp);
	    fclose(fp);
	}
	else
	    pfNotify(PFNFY_WARN, PFNFY_RESOURCE,
		"Could not open scene.out for debug printing.");
    }

    if (loaded)
    {
	pfNotify(PFNFY_INFO,  PFNFY_PRINT, "databases loaded.");
	ViewState->texList = pfuMakeTexList((pfNode *)scene);
    }
    else
    {
    	pfSphere	bsphere;

	pfNotify(PFNFY_INFO, PFNFY_PRINT, "No Databases loaded.");

	/* Set a default bounding sphere */
	pfSetVec3(bsphere.center, 0.0f, 0.0f, 0.0f);
	bsphere.radius = BOUND;
	pfNodeBSphere(scene, &bsphere, PFN_BMODE_DYNAMIC);
    }

    /* Set initial view position and angles */
    initView(scene);
}


/* Initialize graphics pipeline(s) */
void
InitPipe(void)
{
    if (ViewState->input == PFUINPUT_X)
	pfInitPipe(NULL, OpenXPipeline);
    else
	pfInitPipe(NULL, OpenPipeline);

    pfNotify(PFNFY_INFO,  PFNFY_PRINT,
		"Initialized %d Pipe(s)", NumPipes);
}


/* Initialize GUI Panel and widget callback routines */
void
InitGUI(void)
{
    long	masterPipe;

    if (Multipipe)
#if MCCALLUM
    	masterPipe = 0;
#else
    	masterPipe = ChanOrder[(NumChans - 1) >> 1];
#endif
    else
    	masterPipe = 0;

    /* Init GUI and internal data for pipe of master channel */
    pfuInitGUI(pfGetPipe(masterPipe));

    /* Set the initial enable for the GUI */
    pfuEnableGUI(ViewState->gui);
}


/* Initialize Viewing Channel(s) */
void
InitChannel(void)
{
    int		    	i;
    pfFrameStats    	*frameStats;
    pfChannel	    	*chan;
    pfVec3   		xyz, hpr;
    float		fov;

    if (Multipipe)
    {
	/* Distribute channels across pipes as defined by ChanOrder table */
	for (i=0; i< NumChans; i++)
	    Chan[i] = pfNewChan(pfGetPipe(i));
    }
    else
    {
	/* Create all channels on the same pipe */
	for (i=0; i< NumChans; i++)
	    Chan[i] = pfNewChan(pfGetPipe(0));
    }

    /* Master channel is source for keyboard and mouse input */
#if MCCALLUM
    ViewState->masterChan = Chan[0];
#else
    ViewState->masterChan = Chan[ChanOrder[(NumChans - 1) >> 1]];
#endif
    
#if (! MCCALLUM)
    /* If multipipe, then we want channels to share their viewports */
    if (Multipipe)
    {
   	long	share;

	/* Get the default */
	share = pfGetChanShare(ViewState->masterChan);

	/* Add in the viewport share bit */
	share |= PFCHAN_VIEWPORT;

	if (GangDraw)
	{
	    /* Add GangDraw to share mask */
	    share |= PFCHAN_SWAPBUFFERS_HW;
	}
	pfChanShare(ViewState->masterChan, share);
    }
#else
    if (Multipipe)
    {
   	long	share;

	/* Get the default */
	share = pfGetChanShare(ViewState->masterChan);

	/* Add in the viewport share bit */
	share &= ~PFCHAN_VIEW;

#define PERSIA_EARTHSKY_NOT_SHARED 1
#if PERSIA_EARTHSKY_NOT_SHARED
	/* This line will cause earth/sky not to be shared between
	   channels.  It should probably by removed for now until we
	   figure out how to put the sky in all channels and also set
	   this bit.  Note: this line was added so that we wouldn't
	   get fog in the vulture camera view of the highway for my
	   thesis defense video. */
	share &= ~PFCHAN_EARTHSKY;
#endif /* PERSIA_EARTHSKY_NOT_SHARED */

	share &= ~PFCHAN_FOV;

	pfChanShare(ViewState->masterChan, share);
    }
#endif /* !MCCALLUM */

    /* Attach channels to master channel to form a channel group. */
    for (i=0; i< NumChans; i++)
	if (Chan[i] != ViewState->masterChan)
	    pfAttachChan(ViewState->masterChan, Chan[i]);

    /* Make sure we don't wrap field-of-view */
#if ! MCCALLUM
    if (NumChans * 45.0f > 360.0f)
	fov = 360.0f / NumChans;
    else
#endif
    	fov = 45.0f;

#if ! MCCALLUM
    if (Multipipe != HYPERPIPE)
    {
    	/* Set each channel's viewing offset */
    	pfSetVec3(xyz, 0.0f, 0.0f, 0.0f);

	/* Horizontally tile the channels from left to right */
    	for (i=0; i< NumChans; i++)
    	{
	    pfSetVec3(hpr, (((NumChans - 1)*0.5f) - i)*fov, 0.0f, 0.0f);
	    pfChanViewOffsets(Chan[ChanOrder[i]], xyz, hpr);
    	}
    }
#endif

    chan = ViewState->masterChan;

    /* Set the callback routines for the pfChannel */
    pfChanCullFunc(chan, CullFunc);
    pfChanDrawFunc(chan, DrawFunc);

    /* Attach the visual database to the channel */
    pfChanScene(chan, ViewState->scene);

    /* Attach the EarthSky model to the channel */
    pfChanESky(chan, ViewState->eSky);

    /* Initialize the near and far clipping planes */
    pfChanNearFar(chan, ViewState->near, ViewState->far);

    /* Vertical FOV is matched to window aspect ratio. */
    pfChanAutoAspect(chan, PFFRUST_CALC_VERT);
    pfMakeSimpleFrust(chan, fov);

    /* Initialize the viewing position and direction */
    pfChanView(chan, ViewState->initView.xyz, ViewState->initView.hpr);

    /* Configure channel viewports for the Multichannel Option board */
    if (!Multipipe && (NumChans > 1))
	pfuConfigMCO(Chan, NumChans);

    /* Set up Stats structure */
    frameStats = pfGetChanFStats(chan);
    pfStatsAttr(frameStats, PFFSTATS_UPDATE_SECS, 2.0f);
    pfStatsClass(frameStats, PFSTATS_ENGFX, PFSTATS_ON);
    pfStatsClass(frameStats, PFFSTATS_ENDB, PFSTATS_ON);

    /* For GLX windows and X input, get the window info on the master chan */
    if (ViewState->input == PFUINPUT_X)
    {
	pfuGetGLXWin(pfGetChanPipe(ViewState->masterChan), 
		&(ViewState->xWinInfo));
    }
    /* Call custom initGUI routine with the pipe of the master chan */
    initGUI(ViewState->masterChan);

    /* Initialize model for interactive motion */
    initXformer();

    pfNotify(PFNFY_INFO,  PFNFY_PRINT,
	"Initialized %d Channel(s)", NumChans);
}

void
get_world_coords_of_dcs(pfDCS *dcs, pfMatrix mat)
{
  long node_type;
  pfMatrix dcsmat;

  for (pfMakeIdentMat(mat);
       dcs && dcs!=(pfDCS*)ViewState->scene && pfGetNumParents(dcs);
       dcs = (pfDCS*) pfGetParent(dcs, 0))
    {
      node_type = pfGetType(dcs);
      if (node_type == PFTYPE_DCS)
	{
	  pfGetDCSMat((pfDCS*)dcs, dcsmat);
	  pfPostMultMat(mat, dcsmat);
	}
    }
}

/*
 * All latency critical processing should be done here. This is
 * typically the viewing position. One should also read latency critical
 * devices here.
*/
void
PreFrame(void)
{
    /* Poll the mouse buttons and position */
    pfuGetMouse(&ViewState->mouse);

#if MCCALLUM

    /* Call C hook functions */
    {
      int i;
      for (i = 0; i < ViewState->num_c_hooks; i++)
	(*ViewState->c_hooks[i])();
    }

    /* Call Scheme hook functions */
    elk_pre_frame_procedure();

    /* Position all cameras, except 0, the masterChan  */
    {
      int i;
      long node_type;
      pfMatrix mat;
      pfMatrix dcsmat;
      pfGroup *dcs;

      for (i = 0; i < ViewState->num_cameras; i++)
	{
	  if (ViewState->camera_pos[i])
	    {
	      for (dcs = (pfGroup*) ViewState->camera_pos[i], 
		   pfMakeIdentMat(mat);
		   dcs 
		   && dcs!=(pfGroup*)ViewState->scene && pfGetNumParents(dcs);
		   dcs = (pfGroup*) pfGetParent(dcs, 0))
		{
		  node_type = pfGetType(dcs);
		  if (node_type == PFTYPE_DCS)
		    {
		      pfGetDCSMat((pfDCS*)dcs, dcsmat);
		      pfPostMultMat(mat, dcsmat);
		    }
		}
#if PELZ_DEBUGGING
	      printf ("mat %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f\n",
		      mat[0][0], mat[0][1], mat[0][2],
		      mat[1][0], mat[1][1], mat[1][2],
		      mat[2][0], mat[2][1], mat[2][2]);
#endif /* PELZ_DEBUGGING */
	      pfChanViewMat(ViewState->chans[i], mat);
	    }
	  else
	    {
	      if (i == 0)
		updateView(ViewState->masterChan);
	    }
	}
    }
#else /* MCCALLUM */

    updateView(ViewState->masterChan);

#endif /* MCCALLUM */
}


/* Application simulation update is done here */
void
PostFrame(void)
{
    /* Get snapshot of event/key queues */
    pfuGetEvents(&ViewState->events);

    /* Update the simulation state */
    updateSim(ViewState->masterChan);
}


/*************************************************************************
 *									 *
 *		    ISECT PROCESS ROUTINES				 *
 *									 *
 *************************************************************************/

/*
 * Collide the viewing position represented by the pfuXformer with the
 * scene.
*/
void
IsectFunc(void *data)
{
    data;  /* suppress compiler warns */

#if DOING_COLLISION_DETECTION /* Temporarily, to increase frame rate. */
    pfuCollideXformer(ViewState->xformer);
#endif
}


/*************************************************************************
 *									 *
 *		    CULL PROCESS ROUTINES				 *
 *									 *
 *************************************************************************/

/* In Cull callback routine, pre-pfCull() processing */
void
PreCull(pfChannel *chan, void *data)
{
    static long first = 1;
    
    if (first)
    {
	/* 
	 * Force CULL to run on a specific CPU if process locking 
	 * is enabled.
	*/
	if ((pfGetMultiprocess() & PFMP_FORK_CULL) &&
	    (ViewState->procLock & PFMP_FORK_CULL))
		pfuLockDownCull(pfGetChanPipe(chan));
	first = 0;
    }

    (chan, data);	/* To avoid compiler warnings */
    return;
}


/* In Cull callback routine, post-pfCull() processing */
void
PostCull(pfChannel *chan, void *data)
{
    (chan, data);	/* To avoid compiler warnings */
    return;
}


/*****************************************************************************
 *									     *
 *		DRAW PROCESS ROUTINES					     *
 *									     *
 *****************************************************************************/

/* Initialize window */

char WinName[80] = "IRIS Performer";

/* 
 * Create and configure a GLX window.
*/
void
OpenXPipeline(pfPipe * p)
{
    /* -1 -> use default screen or that specified by shell DISPLAY variable */
    long		screen=-1;
    long		xo, yo, xmax, ymax;
    Display 		*XDpy;
    pfuGLXWindow 	*win;

    if (NumPipes > 1)
    {
#if MCCALLUM
	screen = PF_MIN2(pfGetId(p), getgdesc(GD_NSCRNS)-1);
#else
	screen = PF_MIN2(ChanOrder[pfGetId(p)], getgdesc(GD_NSCRNS)-1);
#endif
    }
    XDpy = (Display *) pfuOpenXDisplay(screen);

    xo   = 0;
    yo	 = 0;
#if MCCALLUM
    xmax = xo + 200;
    ymax = yo + 200;
#else
    xmax = xo + (WinSizeX - 1);
    ymax = yo + (WinSizeY - 1);
#endif

    if (!(win = pfuGLXWinopen((pfuXDisplay*) XDpy, p, WinName,
			    xo, xmax, yo, ymax)))
    {
	pfNotify(PFNFY_FATAL, PFNFY_RESOURCE,
	    "OpenXPipeline: Could not create GLX Window.");
    }

    /* initialize Performer for GLX rendering */
    pfInitGLXGfx(p, (void*)XDpy, win->xWin, win->glWin, win->overWin,
		 PFGLX_AUTO_RESIZE);

    /* set initial modes and welcome screen */
    InitGfx(p);
    XFlush(XDpy);

    /* 
     * Force DRAW process to run on a specific CPU for better real-time
     * behavior.
    */
    if (ViewState->procLock & PFMP_FORK_DRAW)
	pfuLockDownDraw(p);
}


/* 
 * Create and configure a GL window.
*/
void
OpenPipeline(pfPipe * p)
{
    long	xo, yo, xmax, ymax;

    if ( NumPipes > 1)
#if MCCALLUM
        scrnselect(PF_MIN2(pfGetId(p), getgdesc(GD_NSCRNS)-1));
#else
        scrnselect(PF_MIN2(ChanOrder[pfGetId(p)], getgdesc(GD_NSCRNS)-1));
#endif

    xo   = 0;
    yo	 = 0;
#if MCCALLUM
    xmax = xo + 200;
    ymax = yo + 200;
#else
    xmax = xo + (WinSizeX - 1);
    ymax = yo + (WinSizeY - 1);
#endif

    /* Do not draw window border if window is full-screen */
    if (WinSizeX == getgdesc(GD_XPMAX) && WinSizeY == getgdesc(GD_YPMAX))
	noborder();

#if MCCALLUM
    {
      int window_number = ViewState->window_num;
      if (ViewState->window_positions[window_number][0] != 0)
	{
	  printf("Placing window at %d %d %d %d\n", 
		 ViewState->window_positions[window_number][0],
		 ViewState->window_positions[window_number][2],
		 ViewState->window_positions[window_number][1],
		 ViewState->window_positions[window_number][3]);
	  
	  prefposition(ViewState->window_positions[window_number][0],
		       ViewState->window_positions[window_number][2],
		       ViewState->window_positions[window_number][1],
		       ViewState->window_positions[window_number][3]);
	}
      else
	prefsize(200, 200);
      (ViewState->window_num)++;
    }
#else
    prefposition(xo, xmax, yo, ymax);
#endif

    foreground();	/* Force GL to be part of this process (no GL fork) */

    winopen(WinName);
    winconstraints();

    /*
     * Call Performer graphics init function to initialize
     * GL window state.
    */
    pfInitGfx(p);

    /* Initialize state of window */
    InitGfx(p);

    /* Queue GL Input devices - KEYBD, MOUSE, queued by default */
    qdevice(F1KEY);

    /* 
     * Force DRAW process to run on a specific CPU for better real-time
     * behavior.
    */
    if (ViewState->procLock & PFMP_FORK_DRAW)
	pfuLockDownDraw(p);
}


/* Initialize graphics state */
void
InitGfx(pfPipe * p)
{
    p;	/* suppress compiler warns */

    /* Set up default state values.  */

    /*
     * Once gfx is initialized (with pfInitGfx), we can check the 
     * Performer environment to see if texture should be on by default
    */
    ViewState->texture = pfGetEnable(PFEN_TEXTURE);
    ViewState->aa = pfGetAntialias();

    if (1 || ViewState->lighting) /* mccallum */
    {
	pfLightModel	*lm;
    	pfMaterial 	*mtl;

	/* 
	 * Set up default material in case the database does not
	 * have any. Turn off lmcolor mode so that database does
	 * not pick up random colors.
	*/
	mtl = pfNewMtl(pfGetSharedArena());
	assert(mtl);
	/* mccallum pfMtlColorMode(mtl, PFMTL_FRONT, LMC_COLOR); */
	pfApplyMtl(mtl);
	/* mtl = pfGetCurMtl(PFMTL_FRONT); */

	/* Set up default lighting model. */
	lm = pfNewLModel(NULL);
#if MCCALLUM
	pfLModelAmbient(lm, 0.7f, 0.7f, 0.7f);
#else
	pfLModelAmbient(lm, 0.0f, 0.0f, 0.0f);
#endif
    	pfApplyLModel(lm);
    }
    /* call custom routine to initialize pipe */
    initPipe(p);
}


/* In Draw callback routine, pre-pfDraw() processing */
void
PreDraw(pfChannel *chan, void *data)
{
    static long		firstTime=1;

    /*
     * Downloading textures here rather than in the OpenPipeline function
     * allows multiple pipelines to download textures in parallel.
    */
    if (firstTime)
    {
    	/* Prebind textures to be used in simulation */
    	pfuDownloadTexList(ViewState->texList, PFUTEX_SHOW);
	/* displaying textures trashes GUI channel so force a redraw */
	if (chan == ViewState->masterChan)
	    pfuRedrawGUI();
	firstTime = 0;
    }

    /* Clear the frame buffer */
    pfClearChan(chan);

    /*
     * If this pipe does not display the GUI, then clear the space where
     * the GUI would ordinarily be.
    */
    if ((Multipipe == 1) && (chan != ViewState->masterChan) && ViewState->gui)
    {
	long	xo, yo, xs, ys;

 	pushviewport();
	pfGetChanOrigin(chan, &xo, &yo);
	pfGetChanSize(chan, &xs, &ys);

	if (ViewState->guiFormat == GUI_VERTICAL)
	    viewport(0, xo-1, 0, ys-1);
	else if (ViewState->guiFormat == GUI_HORIZONTAL)
	    viewport(0, xs-1, 0, yo-1);

	cpack(0xff808080);
	clear();
	popviewport();
    }

    /* Update Draw modes that have been changed in ViewState structure. */
    drawUpdate(chan, data);
}


/* In Draw callback routine, post-pfDraw() processing */
void
PostDraw(pfChannel *chan, void *data)
{
    data;	/* suppress compiler warns */

    /* Display Stats, if active */
    if (ViewState->stats)
	pfDrawChanStats(chan);

    if (ViewState->doing_eye_processing 
	&& chan == ViewState->chans[ViewState->eye_camera_num])  
      eye_processing();		/* defined in eye.c */

    /*
     * Read GL keyboard, mouse. Make sure we only collect input on
     * master channel. If hyperpiping we must collect on all pipes.
    */
    if (chan == ViewState->masterChan)
    	pfuCollectInput();
}
