;;; VR HELMET

;;; Which head-/hand-position sensor hardware are we using?
;;; Currently supported values are the symbols: 'flock and 'polhemus
(if (not (bound? 'ps:*headhad-device*)) (define ps:*headhand-device*
					  'polhemus))

;;; Load the necessary .scm file; this file typically loads the .so file.
(require ps:*headhand-device*)


;;; MATRICES FOR POSITION/ORIENTATION RELATIONSHIPS

;;; A matrix "foo-to-bar" maps points in coordinate system "foo" to
;;; points in coordinate system "bar".  The first row maps foo's x
;;; axis to bar's x-axis.  For example, if foo has x where bar has -y,
;;; then the first row is 0 -1 0.

;;; Another (perhaps inferior) way of thinking about these matrices is
;;; that they describe the position and orientation of one coordinate
;;; system in terms of another coordinate system.  For example
;;; "bar-in-foo" describes the position an orientation of the origin
;;; of bar in terms of foo's coordinate system.  "bar-in-foo" is
;;; equivalent to "foo-to-bar".

;;; The big picture on camera positioning and coordinates: In
;;; generic.c the virtual camera is positioned in global scene
;;; coordinates using pfChanViewMat.  The matrix we want to pass to
;;; this routine is "nose-to-scene".  We obtain this matrix by
;;; beginning at the nose coordinate system, and walking up the scene
;;; tree to the scene coordinate system, post-multiplying pfDCS node
;;; matrices along the way.  For example: nose-to-receiver *
;;; receiver-to-transmitter * transmitter-to-headrest * headrest-to-car *
;;; car-to-ground * ground-to-scene = nose-to-scene.

;;; In this file we are concerned with the "nose-to-headrest" section of
;;; this matrix multiplication sequence.
;;;  nose-to-receiver
;;;   (a constant, this depends on how the receiver is screwed onto the helmet)
;;;  receiver-to-transmitter
;;;   (this one is provided by the polhemus or flock)
;;;  transmitter-to-headrest
;;;   (a constant, depends on how we want world coord system in room)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The relationship between the nose and the receiver
;;; This can be called "nose-to-receiver" or, equivalently,
;;; "receiver-in-nose-coordinate-system"

;;; The "nose" coordinate system has
;;; +x pointing to the right
;;; +y pointing straight ahead
;;; +z pointing up
;;; The origin of the eyes is -.1 meters +z and .04 meters +x
;;; of the receiver

;;; If you hold the polhemus receiver in front of you with the cord
;;;  pointing towards your navel, and the black (flat) side facing the floor,
;;;  then (according to page 22 of the 3SPACE Polhemus User's Manual):
;;; +x points in front of you
;;; +y points to your right
;;; +z points towards the floor

(define ps:*nose-to-headreceiver-pfmatrix*
  (ps:vector->pfmatrix '#(#(0.   1.  0.  0.)
			  #(1.   0.  0.  0.)
			  #(0.   0. -1.  0.)
			  #(0.04 0.  0.1 1.))))

; This value was used for the Flock
;(define ps:*nose-in-headreceiver-pfmatrix*
;  (ps:vector->pfmatrix '#(#(0.   1. 0.   0.)
;			  #(1.   0. 0.   0.)
;			  #(0.   0. -1.  0.)
;			  #(0.2  0. 0.05 1.))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The relationship between the transmitter and the headrest
;;; This can be called "transmitter-to-headrest" or, equivalently
;;; "headrest-in-transmitter-coordinate-system".

;;; The polhemus transimitter is rubber-banded into its mount with
;;; +x pointing towards the wall with blackboard
;;; +y pointing towards the floor
;;; +z pointing towards the robot lab
;;; The origin of the headrest is .5 meters +y from the transmitter, and
;;;  .1 meters +x. 

;;; The headrest coordinate system (when the kart is set up facing the wall
;;;  with the blackboard) is
;;; +x pointing toward the curved wall
;;; +y pointing toward the wall with the blackboard
;;; +z pointing toward the ceiling

(define ps:*headhand-transmitter-to-headrest-pfmatrix*
  (ps:vector->pfmatrix '#(#(0.  1.  0. 0.)
			  #(0.  0. -1. 0.)
			  #(-1. 0.  0. 0.)
			  #(0.1 0.5 0. 1.))))

; This value was used for flock
;(define ps:*headhand-transmitter-in-seat-pfmatrix*
;  (ps:vector->pfmatrix '#(#(1. 0. 0. 0.)
;			  #(0. -1. 0. 0.)
;			  #(0. 0. -1. 0.)
;			  #(0. 1.43 0. 1.))))


(define (ps:place-vr-eyes-in-head-dcs left-camera right-camera eye-sep)
  (let* ((nose-to-headreceiver-dcs (pf:new-dcs))
	 (dcs-eye-left (pf:new-dcs))
	 (dcs-eye-right (pf:new-dcs)))
    (begin
      ;; translate/rotate nose-to-headreceiver-dcs by the correct matrix
      (pf:dcs-mat nose-to-headreceiver-dcs
		  ps:*nose-to-headreceiver-pfmatrix*)
      ;; place the nose in the head
      (pf:add-child ps:*head-dcs* nose-to-headreceiver-dcs)
      ;; translate the left eye to the left of nose
      (pf:dcs-trans dcs-eye-left (- (/ eye-sep 2)) .0 .0)
      ;; translate the right eye to the right of nose
      (pf:dcs-trans dcs-eye-right (/ eye-sep 2) .0 .0)
      ;; place the left eye in/wrt the nose
      (pf:add-child nose-to-headreceiver-dcs dcs-eye-left)
      ;; place the right eye in/wrt the nose
      (pf:add-child nose-to-headreceiver-dcs dcs-eye-right)
      ;; make virtual camera number `left-camera' view from `dcs-eye-left'
      (ps:attach-camera left-camera dcs-eye-left)
      ;; make virtual camera number `right-camera' view from `dcs-eye-right'
      (ps:attach-camera right-camera dcs-eye-right)
      (if (bound? 'ps:*eye-dcs*)
	  (begin
	    ;; place the eye-dcs in the nose
	    (pf:add-child nose-to-headreceiver-dcs ps:*eye-dcs*)
	    ;; place an eye-ray in the eye-dcs
	    (pf:add-child ps:*eye-dcs*
			  (ps:make-geode-cube
			   .1 100. .1
			   (cons 'gstate (ps:make-gstate
					  '(color . #(.5 .5 .1))))))))
      ps:*head-dcs*)))

(define (ps:place-vr-eyes-in-dcs left-camera right-camera eye-sep nose-dcs)
  (let* ((dcs-eye-left (pf:new-dcs))
	 (dcs-eye-right (pf:new-dcs)))
    (begin
      (pf:dcs-trans dcs-eye-left (- (/ eye-sep 2)) .0 .0)
      (pf:dcs-trans dcs-eye-right (/ eye-sep 2) .0 .0)
      (pf:add-child nose-dcs dcs-eye-left)
      (pf:add-child nose-dcs dcs-eye-right)
      (ps:attach-camera left-camera dcs-eye-left)
      (ps:attach-camera right-camera dcs-eye-right)
      (if (bound? 'ps:*eye-dcs*)
	  (begin
	    ;; place the eye-dcs in the nose
	    (pf:add-child nose-dcs ps:*eye-dcs*)
	    ;; place an eye-ray in the eye-dcs
	    (pf:add-child ps:*eye-dcs*
			  (ps:make-geode-cube
			   .1 100. .1
			   (cons 'gstate (ps:make-gstate
					  '(color . #(.5 .5 .1))))))))
      nose-dcs)))

(define (ps:position-windows-for-vr-eyes right-camera-num left-camera-num)
  (ps:position-window right-camera-num (+ 1) (+ 1) 640 480)
  (ps:position-window left-camera-num (+ 1 640) (+ 1) 640 480))

(provide 'headhand)
