/*
iscan.c
Darrell Anderson, Scott Ventura 5/96

see iscan.h for information
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include "persia-scm.h"
#include "persia.h"
#include "iscan.h"

static int iscan_fd;

static pfDCS *iscan_dcs;

static int iscan_active = 0;
static int iscan_initialized = 0;

void iscan_initialize_scm ();

/*----------------------------------------------------------------
  initialize the iscan device (called by P_iscan_Start)
  --------------------------------------------------------------*/
void iscan_initialize(char *device, 
			 short baud, 
			 short timeout )
{
  struct termio term;
  short i;
  unsigned char byte, buf[sizeof(iscan_record)];
  int sync = 0;

  /* open the serial port device */
  if( (iscan_fd = open(device, O_RDWR)) < 0 )
    {
      fprintf(stderr, "failed to open %s for read/write\n", device);
      exit(1);
    }
  
  /* configure term */
  
  /* reset io flag */
  term.c_iflag = 0; 
  /* no output post processing */
  term.c_oflag = 0; 
  /* 8 bit, enable receiver */
  term.c_cflag = (CS8 | CREAD | CLOCAL); 
  /* configure for baud rate */
  switch( baud ) {
  case 2400:
    term.c_cflag |= B2400; break;
  case 4800:
    term.c_cflag |= B4800; break;
  case 9600:
    term.c_cflag |= B9600; break;
  case 19200:
    term.c_cflag |= B19200; break;
  default:
    fprintf(stderr, "illegal baud rate: %d\n", baud);
    exit(1);		
  }
  
  /* no local processing */
  term.c_lflag = 0; 
  term.c_line = (unsigned char) 0;
  
  /* no special chars */
  for( i=0 ; i<NCCS ; i++ ) 
    term.c_cc[i] = (unsigned char)0;
  
  /* set character read timeout */
  term.c_cc[VTIME] = (unsigned char)timeout;
  
  /* minimum characters to read is 0 */
  term.c_cc[VMIN] = 0;
  
  /* configure the device */
  if( (ioctl(iscan_fd, TCSETA, &term)) < 0 ) {
    fprintf(stderr, "failed to set %s characteristics (ioctl)\n", device);
    exit(1);
  }
  
  /* other initializtion/setup */  

#if SYNC_SERIAL
#if 0
  /* iscan records start with 0x4444, so sync up the stream with
     where we expect records to be */
  sync = 0;
  while( !sync ) {
    byte = 0;

    /* find the first 0x44 */  
    while( byte != 0x44 ) {
      while( !iscan_read( &byte, 1 ))
	fprintf(stderr, "failed to read from iscan, continuing.\n");
    }
    
    /* if the next byte is also 0x44, we've read a header */
    while( !iscan_read(&byte, 1))
      fprintf(stderr, "failed to read from iscan, continuing.\n");
    if( byte == 0x44 ) {
      for( i=0 ; i<sizeof(iscan_record)-2 ; i++ ) {
	while( !iscan_read(&byte,1) )
	  fprintf(stderr, "failed to read from iscan, continuing.\n");
      }
      sync = 1;     
    }
  }
#endif
#endif /* SYNC_SERIAL */
}

/*----------------------------------------------------------------
  send numBytes bytes from (*data) to the iscan
  returns 0 on fail, 1 otherwise
  ----------------------------------------------------------------*/
int iscan_send( char *data, short numBytes ) {
  short num;
  
  /* send the data */
  num = write( iscan_fd, data, numBytes );
  
  /* check if we sent it all */
  if( num != numBytes ) {
    fprintf(stderr, "failed to write to iscan device (%d of %d reported)\n",
	    num, numBytes);
    return(0);
  }
  return(1);
}

/*----------------------------------------------------------------
  read numBytes from the iscan, storing it in (*data)
  returns 0 on fail, 1 otherwise
  ----------------------------------------------------------------*/
int iscan_read( char *data, short numBytes ) {
  short num = 0;
  short temp;
  
  /* keep reading until we have enough or something goes wrong */
  while( num < numBytes ) {
    
    /* read what we can */
    temp = read( iscan_fd, data + num, numBytes - num );
    
    /* was there an error? */
    if( temp < 0 ) {
      fprintf(stderr, "failed iscan read (%d bytes reported)\n",
	      temp);
      return(0);
    }

    /* read EOF before reading 'numBytes' bytes.. 
       (not nec. a timeout!) */
#if 0
    if( temp == 0 ) {
      printf("iscan read: EOF after %d of %d, continuing\n", 
	     num, numBytes );
    }
#endif

    /* account for the bytes we did read */
    num += temp;
  }
  return(1);
}

/*----------------------------------------------------------------
  fetch the 'freshest' complete record from the iscan
  ----------------------------------------------------------------*/
void iscan_fetch_record( iscan_record *data ) {
  static iscan_record rec[2];
  static int now_reading = 0;
  static int now_reading_offset = 0;
  int valid_record;
  int amount_read;
  static int how_stale = 0;

  how_stale++;
  if( how_stale >= 100 ) {
    printf( "Warning: iscan record has not updated in a while.  Check the iscan is on and (M)ove has been selected.\n" );
  }

  amount_read = 
    read( iscan_fd, ((unsigned char *)(&(rec[now_reading]))) + 
	  now_reading_offset, sizeof(iscan_record) - now_reading_offset );
  
  while( amount_read == sizeof(iscan_record) - now_reading_offset ) {
    how_stale = 0;

    now_reading_offset = 0;
    now_reading = (now_reading+1)&0x01;
    
    amount_read = 
      read( iscan_fd, ((unsigned char *)(&(rec[now_reading]))) + 
	    now_reading_offset, sizeof(iscan_record) - now_reading_offset );    
  }
  now_reading_offset += amount_read;

  valid_record=(now_reading+1)&0x01;    
  memcpy( data, &(rec[valid_record]), sizeof(iscan_record) ); 
}


/*----------------------------------------------------------------
  update global matrices, called as a pre-frame hook [*]
  --------------------------------------------------------------*/
void iscan_c_hook(void)
{
  iscan_record data;
  pfMatrix euler_matrix;
  float hf, vf;
  /* for sync code */
  unsigned char byte;
  int i, sync = 0;

/* iscan_read( &data, sizeof(iscan_record) ); */
  iscan_fetch_record( &data );    
  
#if SYNC_SERIAL
  if( data.header[0] != 0x44 || data.header[1] != 0x44 ) {

#if 0
    /* some information on what has been read */
    printf( "\n[0x" );
    for( i=0 ; i<sizeof(iscan_record) ; i++ )
      printf( "_%x", *(((unsigned char *)&data)+i) );
    printf( "] " );
    printf( "['" );
    for( i=0 ; i<sizeof(iscan_record) ; i++ )
      printf( "%c", *(((unsigned char *)&data)+i) );
    printf( "'] " );
    printf( "warning: iscan data header is not 0x4444.\n");
#endif

    while( !sync ) {
      byte = 0;      
      /* find the first 0x44 */  
      while( byte != 0x44 ) {
	while( !iscan_read( &byte, 1 ))
	  printf( "*");
      }
#if 0
      printf( "read first header byte: 0x%x\n", byte );
#endif

      /* if the next byte is also 0x44, we've read a header */
      while( !iscan_read(&byte, 1))
	printf( "*");
#if 0
      printf( "read second header byte: 0x%x\n", byte );
#endif

      if( byte == 0x44 ) {
#if 0
	for( i=0 ; i<(sizeof(iscan_record)-2) ; i++ ) {
	  while( !iscan_read(&byte,1) )
	    printf( "*" );
	  printf( "dropped 0x%x\n", byte );
	}
#else
	iscan_read((unsigned char *)&(data.h),2);
	iscan_read((unsigned char *)&(data.v),2);
#endif
	sync = 1;
      }
    }
  } 
#endif /* SYNC_SERIAL */

  reverse_byte_order2( (unsigned char *) &(data.h) );
  reverse_byte_order2( (unsigned char *) &(data.v) );

#if 1 
  printf("[%hd,%hd]\n",data.h,data.v);
#endif

  /* Put the rotation into the matrix. */
  hf = data.h;
  vf = data.v;
  pfMakeEulerMat (euler_matrix, hf, vf, 0);
  /* Put the matrix into the DCS. */
  pfDCSMat (iscan_dcs, euler_matrix);
}

/*----------------------------------------------------------------
  master iscan initialization (and activation) funtion, 
  called from scheme [*]
  ----------------------------------------------------------------*/
Object P_iscan_Start(void) {
  if( !ViewState ) 
    Primitive_Error("ps:init-cameras not yet called");
  
  iscan_initialize_scm ();
  /* do this each time the Polhemus is re-enabled */
  if( !iscan_active ) {
    iscan_active = 1;
    persia_add_pre_frame_hook(iscan_c_hook);
  }
  
  /* do this only once! */
  if( !iscan_initialized ) {
    iscan_initialized = 1;
    iscan_initialize(ISCAN_DEVICE, ISCAN_BAUD, ISCAN_TIMEOUT );
  }
  return Void;
}

/*----------------------------------------------------------------
  master iscan kill(suspend really) function, called from scheme [*]
  ----------------------------------------------------------------*/
Object P_iscan_Stop(void) {
  if( !ViewState ) 
    Primitive_Error("ps:init-cameras not yet called");
  /* disable the iscan */
  if( iscan_active ) {
    iscan_active = 0;
    persia_remove_pre_frame_hook(iscan_c_hook);
  }
  return Void;
}

/*----------------------------------------------------------------
  make above scheme functions available (a la Define_Primitive) [*]
  ----------------------------------------------------------------*/
void iscan_initialize_scm ()
{
  static int done = 0;
  if (!done)
    {
      done = 1;
      /* create the structures */
      iscan_dcs = pfNewDCS();
      /* make them available in scheme */
      Define_Pfo((pfObject*)iscan_dcs, "ps:*eye-dcs*");
    }
}

void reverse_byte_order2( unsigned char *word ) {
  unsigned char temp;

  temp = word[0];
  word[0] = word[1];
  word[1] = temp;
}

void init_iscan(void)
{
  /* make the functions available in scheme */
  Define_Primitive(P_iscan_Start, "ps:iscan-start", 0, 0, EVAL);
  Define_Primitive(P_iscan_Stop, "ps:iscan-stop", 0, 0, EVAL);
}


