/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/*
 * keybd.c -- keyboard commands
 */

#include <stdio.h>
#include "pfutil.h"

#include "perfly.h"
#include "gui.h"


void 
printCommands(void)
{
    fprintf(stderr, " Key     Function\n");
    fprintf(stderr, "-----    --------------\n");
    fprintf(stderr, " ESC     Exit program \n");
    fprintf(stderr, "  ?      Print help\n");
    fprintf(stderr, "SPACE    Stop moving\n");
    fprintf(stderr, "F1KEY    Toggle graphical user interface\n");
    fprintf(stderr, "  b      Toggle backface mode\n");
    fprintf(stderr, "  c      Change culling mode\n");
    fprintf(stderr, "  f      Change fields per frame\n");
    fprintf(stderr, "  g      toggle stats and stats graph\n");
    fprintf(stderr, "  G      print stats graph contents to stderr\n");
    fprintf(stderr, "  l      Toggle lighting\n");
    fprintf(stderr, "  p      Change frame phase\n");
    fprintf(stderr, "  r      Reset positions\n");
    fprintf(stderr, "  s      Toggle diagnostics display\n");
    fprintf(stderr, "  S      print stats graph contents to stderr\n");
    fprintf(stderr, "  t      Toggle texturing\n");
    fprintf(stderr, "  w      Toggle wireframe\n");
    fprintf(stderr, "  x      Snapshot screen image to file\n");
    fflush(stderr);
}

void 
toggleBackface(void)
{
    ViewState->backface = !ViewState->backface;
    drawModesChanged();
}


void 
toggleWireframe(void)
{
    ViewState->wireframe = !ViewState->wireframe;
    drawModesChanged();

    /* The GUI Controls this mode also, so update the Widget */
    updateWidget(GUI_WIREFRAME, (float)ViewState->wireframe);
}


void 
toggleLighting(void)
{
    ViewState->lighting = !ViewState->lighting;
    drawModesChanged();

    /* The GUI Controls this mode also, so update the Widget */
    updateWidget(GUI_LIGHTING, (float)ViewState->lighting);
}


void 
toggleTexture(void)
{
    ViewState->texture = !ViewState->texture;
    drawModesChanged();

    /* The GUI Controls this mode also, so update the Widget */
    updateWidget(GUI_TEXTURE, (float)ViewState->texture);
}


void 
toggleStats(void)
{
    ViewState->stats = !ViewState->stats;
    /* The GUI Controls this mode also, so update the Widget */
    updateWidget(GUI_STATS,(float)ViewState->stats);
}


void 
changeFieldRate(int inc)
{
    long fields, vrate;

    vrate = pfGetVideoRate();

    fields =  pfGetFieldRate();		/* Get current field rate */

    fields += inc;

    /* Clamp fields */
    if (fields <= 0)
	fields = 1;
    if (fields > vrate)
	fields = vrate;

    pfFieldRate(fields);

    ViewState->frameRate = pfGetFrameRate();	/* Get current frame rate */
}


void
toggleCullMode(void)
{
    long	oldMode	= 0;
    long	newMode	= 0;

    /* get old cull mode */
    oldMode = ViewState->cullMode;

    /* advance to next feature combination */
    newMode = ((oldMode + 1) & 0x0f) | (oldMode & ~0x0f);

    /* GeoSet culling only makes sense when view culling is enabled */
    if (newMode & PFCULL_GSET) newMode |= PFCULL_VIEW;

    /* set new cull mode */
    ViewState->cullMode = newMode;
    
    updateWidget(GUI_CULL, newMode);
}

void
togglePhase(void)
{
    pfPhase((pfGetPhase() + 1) % 4);
}
			

void 
toggleGUI(void)
{
    ViewState->gui = !ViewState->gui;
    ViewState->updateChannels = 1;
}


/* process input in the app process space */
void
processKeybdInput(void)
{
    long 	 i;
    long	 j;
    long	 key;
    long 	 dev;
    long	 val;
    long	 numDevs;
    pfuEvent 	*events;
    
    events = &ViewState->events;
    numDevs = events->numDevs;

    for (j = 0; j < numDevs; j++)
    {
	dev = events->devQ[j];
	val = events->devVal[j];

	if (events->devCount[dev] > 0)
	{
	    switch(dev)
	    {
	    /* Window-system events */
	    case PFUDEV_REDRAW:
		events->devCount[dev] = 0; /* mark device done */
		pfuRedrawGUI();
		ViewState->redrawOverlay = 3;
		for (i=0; i<NumPipes; i++)
		    ViewState->drawFlags[i] |= REDRAW_WINDOW;
		break;
		
	    case PFUDEV_WINQUIT:
		ViewState->exitFlag = 1;
		events->devCount[dev] = 0; /* mark device done */
		break;
		
	    /* Main keyboard */
	    case PFUDEV_KEYBD:
		for (i = 0; i < events->numKeys; i++)
		{
		    key = events->keyQ[i];
		    if (events->keyCount[key])
		    {
			switch(key)
			{
			case 27:	/* ESC  exit program */
			    ViewState->exitFlag = 1;
			    break;

			case '?':	/* print commands */
			    printCommands();
			    break;
		
			case ' ':	/* Stop moving */
			    pfuStopXformer(ViewState->xformer);
			    break;
		
			case 'b':	/* Toggle backface mode */
			case 'B':
			    toggleBackface();
			    break;

			case 'c':	/* Change culling mode */
			case 'C':
			    toggleCullMode();
			    break;

			case 'f':	/* Increment fields per frame */
			    changeFieldRate(1);
			    break;

			case 'F':	/* Decrement fields per frame */
    			    changeFieldRate(-1);
			    break;

			case 'l':	/* Toggle lighting */
			case 'L':
			    toggleLighting();
			    break;

			case 'p':	/* Change frame phase */
			case 'P':
			    togglePhase();
			    break; 

			case 'r':	/* Reset positions */
			case 'R':
			    resetPosition(POS_ORIG);
			    break;
			
			case 's':	/* Toggle diagnostics display */
			case 'g':
			    toggleStats();
			    break;
			
			case 'S':	/* Print stats display to stderr */
			case 'G':
			    if (ViewState->stats)
				ViewState->printStats = 1;
			    break;
			    
			case 't':	/* Toggle texturing */
			case 'T':
			    toggleTexture();
			    break;
	
			case 'w':	/* Toggle wireframe */
			case 'W':
			    toggleWireframe();
			    break;
			    
			case 'x':	/* Snap screen image */
			case 'X':
			    ViewState->snapImage = 1;
			    break;

			default:	/* Unsupported keyboard input */ 
			    break;
			} 			
			/* reset the key as done */
			events->keyCount[key] = 0;
		    } 		    
		    events->numKeys = 0;
		} 
		/* mark device done */
		events->devCount[dev] = 0; 
		break; 
		
	    /* Function keys */
	    case PFUDEV_F1KEY:		/* GUI on / off */
		toggleGUI();
		/* mark device done */
		events->devCount[dev] = 0; 
		break;
	    }
	}
    }
    events->numDevs = 0;
}
