/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/*
 *	main.c
*/

#include <stdlib.h>
#include <stdio.h>

/* proceess control includes */
#include <sys/statfs.h>
#include <sys/types.h>


/* Performer include files */
#include <Performer/pf.h>
#include "pfutil.h"

/* Common Code include file */
#include "generic.h"

/* 'custom.h' is provided by the individual applications. It must contain
    the typedef for the ViewState structure
*/
#include "custom.h"
#include <sys/types.h>
#include <sys/time.h>

/**************************************************************************
 *									  *
 *		    APPLICATION PROCESS ROUTINES			  *
 *									  * 
 **************************************************************************/


/* frees any restricted/isolated CPUs when app starts up */
long FreeInitCPUs = 1;

void
pre_start_graphics()
{
  abort();			/* not used */
    /* 
     * Initialize IRIS Performer: create shared memory, initialize clocks, etc.
     * This must be the first libpf call in a libpf application.
     */
    pfInit();

    /*
     * Malloc structures now, before pfConfig(), so forked processes have same
     * pointers.
     */
    /* Do this before start_graphics, so that elk has a place to
       put the window position preferences */
    ViewState = (SharedViewState *) pfCalloc(1, sizeof(SharedViewState), 
					     pfGetSharedArena());

}

void
start_graphics(int argc, char *argv[], int num_cameras)
{
  pfInit();

    /* 
     * Allocate and initialize global shared memory before pfConfig 
     * so that forked processes share global pointers. 
     */
    InitSharedMem(argc, argv, num_cameras);

    /* Set multiprocessing mode and other configuration values. */
    InitConfig();

    /* 
     * Configure IRIS Performer and fork extra processes if 
     * configured for multiprocessing. 
     */
    pfConfig();

    /* 
     * Assign a non-degrading priority to all Performer processes. 
     * User processes spawned from this one will inherit the same priority.
    */
    if (PrioritizeProcs)
	pfuPrioritizeProcs(TRUE);
   
    /* Initialize IRIS Performer utility library. */
    pfuInitUtil();
       
    /* Create the visual database. */
    InitScene();

    /* Initialize the rendering pipeline(s). */
    InitPipe();
    
    /* 
     * Initialize the graphical user interface (GUI). 
     * Do this before InitChannel so that the GUI is the first channel.
    */
    InitGUI();

    /* Initialize pfChannel(s). */
    InitChannel();

    /* Initialize input handling (X or GL) for mouse and event inputs. */
    pfuInitInput(pfGetChanPipe(ViewState->masterChan), ViewState->input);
 
    /* 
     * Set up processor locking for APP - must be done after pfConfig() 
     * and should be done after InitInput which forks off an input handling
     * process.
    */
    if (ViewState->procLock == ((ulong)~0))
	pfuLockDownApp();
    
    /* Set the desired frame rate. */
    pfFrameRate(ViewState->frameRate);	

    /* Set the MP synchronization phase. */
    pfPhase(ViewState->phase);		
}

void
one_loop()
{
  /* Sleep until next frame */
  pfSync();
  
  /* 
   * Should do all latency critical processing inbetween pfSync()
   * and pfFrame(). Typically this is viewing position update.
   */
  PreFrame();
  
  /* Trigger cull and draw processing for this frame. */
#if PELZ_DEBUGGING
  {
    int frame_count;
    frame_count = pfFrame();
    printf ("pfFrame frame count %d\n", frame_count);
  }
#else
  pfFrame();
#endif /* PELZ_DEBUGGING */

  /* Increment the frame count */
  (ViewState->frame_count)++;
  /* fprintf(stderr, "main frame count %d\n", ViewState->frame_count); */
  
  /* Perform non-latency critical simulation updates. */
  PostFrame();
}

void
run_loop()
{
  struct timeval tv1, tv2;
  struct timezone tz;
  long loopCount = 0;

  ViewState->exitFlag = 0;

  gettimeofday(&tv1, &tz);
  while (!ViewState->exitFlag)
    {
      loopCount++;
      one_loop();
    }
  gettimeofday(&tv2, &tz);
  printf("%ld frames, %f frames per second\n", loopCount,
	 (float)loopCount / (float)(tv2.tv_sec - tv1.tv_sec));
}

void
cleanup_graphics()
{
    /* 
     * Cleanup and exit.  
     */
    pfuFreeCPUs();
    
    pfuExitUtil();

    pfuExitInput();

    pfExit();
}

int
main (int argc, char *argv[])
{
/*  
  start_graphics(argc, argv, 2);
  run_loop(); */

  /* pre_start_graphics(); */

  Elk_Init(argc, argv, 1, "toplevel.scm");
  cleanup_graphics();
  exit(0);
}


/**************************************************************************
 *									  *
 *		    CULL PROCESS ROUTINES				  *
 *									  * 
 **************************************************************************/

/*
 * The cull function callback.  Any work that needs to be done
 * in the cull process should happen in this function.
*/
void
CullFunc(pfChannel * chan, void *data)
{
    PreCull(chan, data);

    pfCull();			/* Cull to the viewing frustum */

    PostCull(chan, data);
}


/**************************************************************************
 *									  *
 *		    DRAW PROCESS ROUTINES				  *
 *									  * 
 **************************************************************************/

/*
 * The draw function callback.  I/O with GL devices must happen here.
 * Any GL functionality outside Performer must be done here. 
 * All libpr routines which may be captured by a pfDispList, 
 * e.g.- pfEnable, pfApply*, pfCullFace, pfPushMatrix, should be called 
 * here since they send GL commands to the graphics hardware.
*/
void
DrawFunc(pfChannel *chan, void *data)
{
    PreDraw(chan, data);	/* e.g. - clear the viewport */

    pfDraw();			/* Render the frame */

    PostDraw(chan, data);	/* e.g. - draw HUD, read GL devices */
}
