/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/*
 *	perfly.c -- The Performer database fly-thru application
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <bstring.h>
#include <math.h>
#include <ctype.h>
#include <gl/device.h>

#include <Performer/pf.h>
#include <Performer/prmath.h>
#include "pfutil.h"
#include "pfflt.h"
#include "pfsgi.h"

#include "generic.h"

#include "perfly.h"
#include "gui.h"
#include "keybd.h"

#if MCCALLUM
#include "limits.h"
#include "elk.h"
#endif

static void updateStats(pfChannel*);
static void drawOverlayName(void);

	/*----------------------------------------------------*/

/*
 * Initialize ViewState shared memory structure 
 */
void 
initViewState(int num_cameras)
{
    char	gname[12];
    long	i;

    /* Is this a Reality Engine system ? */
    gversion(gname);
    ViewState->realityGraphics = !strncmp("GL4DRE", gname, 6);
 
    /* Set up defaults */
    ViewState->exitFlag		= FALSE;
    ViewState->updateChannels	= TRUE;
#if MCCALLUM
    ViewState->gui		= FALSE;
#else
    ViewState->gui		= TRUE;
#endif
    ViewState->guiFormat	= 0; /* marks unset */
#if PELZ_DEBUGGING
    /* Without this we can get delay of one extra frame between Polhemus/Kart
       and rendered image.  ...No, we always get it? */
    ViewState->phase            = PFPHASE_LIMIT;
#else
    ViewState->phase		= PFPHASE_FREE_RUN;
#endif
#if PELZ_DEBUGGING
    ViewState->frameRate	= 1.0f;
#else
    ViewState->frameRate	= 30.0f;
#endif
    ViewState->aa		= PF_OFF;
    ViewState->backface		= PF_ON;
    ViewState->collideMode	= PF_ON;
    ViewState->earthSkyMode	= PFES_SKY_GRND;
    ViewState->fade		= ViewState->realityGraphics;
    ViewState->fadeRange	= 100.0f;
#if MCCALLUM
    ViewState->near		= 0.01f;
#else
    ViewState->near		= 1.0f;
#endif
    ViewState->far		= 16000.0f;
    ViewState->farFogRange	= ViewState->far;
    ViewState->fog		= PFFOG_OFF;
    ViewState->highLoad		= 0.8f;
    ViewState->lighting		= PF_ON;
    ViewState->lowLoad		= 0.5f;
    ViewState->nearFogRange	= ViewState->near;
    ViewState->stats		= PF_OFF;
    ViewState->stress		= PF_OFF;
    ViewState->stressMax	= 100.0f;
    ViewState->stressScale	= 1.0f;
    ViewState->wireframe	= PF_OFF;
    ViewState->xformerMode	= PFUXF_TRACKBALL;
    ViewState->input		= PFUINPUT_GL;
    ViewState->explode		= 0.0f;
    for (i=0; i < MAX_PIPES; i++)
	ViewState->drawFlags[i] = REDRAW_WINDOW;
#if MCCALLUM
    ViewState->redrawOverlay	= 0;
#else
    ViewState->redrawOverlay	= 3;
#endif
    ViewState->procLock		= 0;
    ViewState->cullMode		= PFCULL_ALL;
    ViewState->lamps		= 0;
#if MCCALLUM
    ViewState->LODscale		= 0.2f;
#else
    ViewState->LODscale		= 1.0f;
#endif

#if MCCALLUM
    ViewState->num_cameras = num_cameras;
    {
      int i;
      for (i = 0; i < MAX_CHANS; i++)
	{
#if 0
	  printf("elk %d %d %d %d\n",
		 elk_window_positions[i][0],
		 elk_window_positions[i][1],
		 elk_window_positions[i][2],
		 elk_window_positions[i][3]);
#endif
	  ViewState->window_positions[i][0] = elk_window_positions[i][0];
	  ViewState->window_positions[i][1] = elk_window_positions[i][1];
	  ViewState->window_positions[i][2] = elk_window_positions[i][2];
	  ViewState->window_positions[i][3] = elk_window_positions[i][3];
	}
    }

    ViewState->num_c_hooks = 0;

    ViewState->frame_count = 0;
#endif /* MCCALLUM */
}

/*
 * Initialize the view
 */
void 
initView(pfScene *scene)
{
    pfSphere	bsphere;
        
    pfGetNodeBSphere(scene, &bsphere);

    /* Set initial view position from command line values */
    if (InitXYZ) 
        pfCopyVec3(ViewState->viewCoord.xyz, ViewState->initView.xyz);

    /* Otherwise fit view to encompass scene */
    else	
    {
        pfCopyVec3(ViewState->viewCoord.xyz, bsphere.center);

        /* Offset view so all visible */
        ViewState->viewCoord.xyz[PF_Y] -= ViewState->sceneSize;

        /* Store initial XYZ position for future reset actions */
        pfCopyVec3(ViewState->initView.xyz, ViewState->viewCoord.xyz);
    }
 
#ifndef MCCALLUM
    /* Calculate near, far based on scene extent */
    ViewState->near = PF_MIN2(ViewState->sceneSize / 10.0f, 1.0f);
    ViewState->far = PF_MAX2(ViewState->sceneSize * 2.0f, 16000.0f);
#endif

    /* Set initial view angles from command line values ? */
    if (InitHPR)
        pfCopyVec3(ViewState->viewCoord.hpr, ViewState->initView.hpr);
    else
    {
	pfSetVec3(ViewState->viewCoord.hpr, 0.0f, 0.0f, 0.0f);

	/* Store initial HPR angles for future reset actions */
	pfCopyVec3(ViewState->initView.hpr, ViewState->viewCoord.hpr);
    }
}


long
initSceneGraph(pfScene *scene)
{
    pfGroup     *root;
    long	i;
    long	loaded	= 0;
    pfSphere    bsphere;

    /* Create a DCS for TRACKBALL pfuXformer */
    ViewState->sceneDCS = pfNewDCS();
    ViewState->sceneGroup = pfNewGroup();
    pfAddChild(scene, ViewState->sceneDCS);
    pfAddChild(ViewState->sceneDCS, ViewState->sceneGroup);

    /* Load each of the files named on the command line */
    for (i = 0;	i < NumFiles; i++)
    {
	/* Load the database. create a hierarchy under node "root" */
	root = (pfGroup *)LoadFile(DatabaseFiles[i], NULL);
	    
	if (root == NULL)
	  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, 
	      "WARNING: could not load \"%s\"", DatabaseFiles[i]);
	else
	{
	    /* explode the object (move it from the origin) */
	    if (ViewState->explode != 0.0f)
	    {
		float	 x;
		float	 y;
		float	 z;
		pfMatrix matrix;
		pfSCS	*scs;

		x = ViewState->explode*((random() & 0xffff)/65535.0f - 0.5f);
		y = ViewState->explode*((random() & 0xffff)/65535.0f - 0.5f);
		z = ViewState->explode*((random() & 0xffff)/65535.0f - 0.5f);

		/* move the object by placing it beneath an SCS node */
		pfMakeTransMat(matrix, x, y, z);
		scs = pfNewSCS(matrix);
		pfAddChild(scs, root);

		/* go ahead and apply the SCS matrix to the geometry */
		pfFlatten(scs, 0);

		/* detach (transformed) object from SCS; discard SCS */
		pfRemoveChild(scs, root);
		pfDelete(scs);
	    }

	    /* Add the hierarchy under the DCS so it can be transformed */
	    pfAddChild(ViewState->sceneGroup, root);
	    ++loaded;
	}
    }

    if (loaded)
    {
    	/* Compute extent of scene */
    	pfGetNodeBSphere(scene, &bsphere);
    	ViewState->sceneSize = PF_MIN2(2.5f*bsphere.radius, 40000.0f);
    }
    else
    	ViewState->sceneSize = 200.0f;
    
    /* add light sources to the scene */
    if (ViewState->lamps > 0)
    {
	for (i = 0; i < ViewState->lamps; i++)
	{
	    pfLightSource *lamp = pfNewLSource(); 

	    /* set direction of infinite light source */
	    pfNormalizeVec3(ViewState->lampXyz[i]);
	    pfLightPos(lamp, 
	        ViewState->lampXyz[i][0], 
	        ViewState->lampXyz[i][1],
	        ViewState->lampXyz[i][2],
	        0.0f);

	    /* set light source color */
	    pfLightColor(lamp, 
	        ViewState->lampRgb[i][0], 
	        ViewState->lampRgb[i][1],
	        ViewState->lampRgb[i][2]);

	    /* add lamp to scene */
	    pfAddChild(scene, lamp);
	}
    }

    return(loaded);
}


/* 
 * Update the current view 
 */
void
updateView(pfChannel *chan)
{
    pfMatrix 	mat;

    pfuUpdateXformer(ViewState->xformer);

    switch (pfuGetXformerMode(ViewState->xformer))
    {
    case PFUXF_FLY:
    case PFUXF_DRIVE:
	pfuGetXformerCoord(ViewState->xformer, &ViewState->viewCoord);
	pfuGetXformerMat(ViewState->xformer, mat);
	pfChanViewMat(chan, mat);
	break;

    case PFUXF_TRACKBALL:
	pfuGetXformerMat(ViewState->xformer, mat);
	pfDCSMat(ViewState->sceneDCS, mat);
	break;
    }
}


/* 
 * Reset view to original XYZ position and HPR Euler angles 
 */
void 
resetPosition(long pos)
{
    pfMatrix	mat;
    pfSphere    bsphere;

    switch (pos)
    {
	case POS_ORIG:
	    ViewState->viewCoord = ViewState->initView;
	    break;
	case POS_CENTER:
	    pfGetNodeBSphere(ViewState->scene, &bsphere);
	    PFCOPY_VEC3(ViewState->viewCoord.xyz, bsphere.center);
	    PFCOPY_VEC3(ViewState->viewCoord.hpr, ViewState->initView.hpr);
	    break;
    }
    

    switch (pfuGetXformerMode(ViewState->xformer))
    {
    case PFUXF_FLY:
    case PFUXF_DRIVE:
	pfuXformerCoord(ViewState->xformer, &ViewState->viewCoord);
	pfMakeIdentMat(mat);
	pfDCSMat(ViewState->sceneDCS, mat);
	break;	

    case PFUXF_TRACKBALL:
	pfuXformerCoord(ViewState->xformer, &ViewState->viewCoord);
	pfuGetXformerMat(ViewState->xformer, mat);
	pfChanViewMat(ViewState->masterChan, mat);
	pfMakeIdentMat(mat);
	pfuXformerMat(ViewState->xformer, mat);
	pfDCSMat(ViewState->sceneDCS, mat);
	break;
    }

    pfuStopXformer(ViewState->xformer);
}


/*
 * Perform non-latency-critical per-frame updates
 */ 
void
updateSim(pfChannel *chan)
{
    /* Process keyboard input */
    processKeybdInput();
	
    /* Adjust load management filter */
    if (ViewState->stress)
	pfChanStressFilter(chan, 1.0f, ViewState->lowLoad, ViewState->highLoad,
	    ViewState->stressScale, ViewState->stressMax);
    else
	pfChanStressFilter(chan,1.0f, ViewState->lowLoad, ViewState->highLoad, 
	    0.0f, 1.0f);

    if (ViewState->gui)
	updateGUIViewMsg();

    if (ViewState->updateChannels)
    {
	updateGUI(ViewState->gui);
	ViewState->updateChannels = 0;
    } 
    else if (ViewState->gui)
	pfuUpdateGUI();

    if ((ViewState->cullMode != -1)&&
	    (ViewState->cullMode != 
	     pfGetChanTravMode(ViewState->masterChan, PFTRAV_CULL)))
	pfChanTravMode(ViewState->masterChan, PFTRAV_CULL, ViewState->cullMode);

    if (ViewState->LODscale != pfGetChanLODAttr(ViewState->masterChan, PFLOD_SCALE))
	pfChanLODAttr(ViewState->masterChan, PFLOD_SCALE, ViewState->LODscale);
	
    updateStats(chan);
    updateEnvironment();
}

static void
updateStats(pfChannel *chan)
{
    /*
     * Note: if turning on or off fill stats, we must redraw panel because
     * multisampling might be toggled as well. 
    */
    static long     on = -1, en = 0;

    pfFrameStats   *fsp;

    if (on && (!ViewState->stats))	/* Turn off all stats collection */
    {
	fsp = pfGetChanFStats(chan);
	pfStatsClass(fsp, PFSTATS_ALL, 0);
	on = 0;
	en = 0;
    } 
    else if (ViewState->stats && (en != ViewState->statsEnable))
    {
	/* Set new stats enables */
	fsp = pfGetChanFStats(chan);

	/* First turn off old stats */
	if (en)
	    pfStatsClass(fsp, en, 0);

	/* Now enable new stats graph */
	if (ViewState->statsEnable)
	{
	    static long     first = 1;
	    pfStatsClass(fsp, ViewState->statsEnable, 1);
	    if (first)
	    {	/* set some stats class modes that are not
		 * on by default. this only needs to be done
		 * the first time */
		pfStatsClassMode(fsp, PFSTATSHW_GFXPIPE_FILL,
				 PFSTATSHW_GFXPIPE_FILL_DEPTHCMP | 
				 PFSTATSHW_GFXPIPE_FILL_TRANSPARENT,
				 PFSTATS_ON);
		/*
		 * This will also enable the tmesh statistics - not on by
		 * default because they are somewhat expensive 
		 */
		pfStatsClassMode(fsp, PFSTATS_GFX,
				 PFSTATS_GFX_ATTR_COUNTS |
				 PFSTATS_GFX_TSTRIP_LENGTHS,
				 PFSTATS_ON);
		pfStatsAttr(fsp, PFFSTATS_UPDATE_SECS, 2.0f);
		first = 0;
	    }
	}
	en = ViewState->statsEnable;
	on = 1;
    }
    if (ViewState->printStats)
    {
	fsp = pfGetChanFStats(chan);
	pfPrint(pfGetChanFStats(ViewState->masterChan), 
		PFFSTATS_BUF_AVG | PFFSTATS_BUF_PREV | 
		pfGetStatsClass(fsp, PFSTATS_ALL), 
		PFPRINT_VB_ON, 0);

	ViewState->printStats = 0;
    }
}


void
xformerMode(long mode, long collideMode)
{
    pfMatrix		mat;
    static long		oldMode = -1, oldCollMode = -1;

    if ((mode == oldMode && oldCollMode == collideMode) || !ViewState->xformer)
	return;

    pfuXformerMode(ViewState->xformer, mode);

    /* If not in trackball mode, remove trackball DCS for better performance */
    if ((oldMode != mode) && (oldMode == PFUXF_TRACKBALL))
    {
	pfRemoveChild(ViewState->scene, ViewState->sceneDCS); 
	pfAddChild(ViewState->scene, ViewState->sceneGroup); 
    }
    /* Restore trackball DCS to scene */
    else if ((oldMode != mode) && (mode == PFUXF_TRACKBALL)) 
    {
	pfRemoveChild(ViewState->scene, ViewState->sceneGroup); 
	if (pfGetNumParents(ViewState->sceneDCS) == 0)
	    pfAddChild(ViewState->scene, ViewState->sceneDCS); 
    }
    
    switch (mode) 
    {
    case PFUXF_DRIVE:
	/* Limit to ~60mph, 90 degrees/sec pan, .125G acceleration */
	pfuXformerLimits(ViewState->xformer, 30.0f, 90.0f, 1.25f, NULL);

	/* Follow 2 meters above ground */
	if (collideMode == PF_ON)
	    pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_GROUND, 2.0f, 
		ViewState->scene);
	else
	    pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_GROUND, 0.0f, 
		NULL);
	
	/* Collide with objects in scene */
	if (collideMode == PF_ON)
	    pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_OBJECT, 
		2.0f*ViewState->near, ViewState->scene);
	else
	    pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_OBJECT, 0.0f, 
		NULL);

	pfuXformerCoord(ViewState->xformer, &ViewState->viewCoord);
	break;

    case PFUXF_FLY:
	/* Limit to ~180mph, 90 degrees/sec pan, .25G acceleration */
	pfuXformerLimits(ViewState->xformer, 90.0f, 90.0f, 2.5f, NULL);

	/* Do not terrain follow */
	pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_GROUND, 0.0f, NULL);

	/* Collide with objects in scene */
	if (collideMode == PF_ON)
	    pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_OBJECT, 
		2.0f*ViewState->near, ViewState->scene);
	else
	    pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_OBJECT, 0.0f, 
		NULL);

	pfuXformerCoord(ViewState->xformer, &ViewState->viewCoord);
	break;

    case PFUXF_TRACKBALL:

	pfuXformerLimits(ViewState->xformer, ViewState->sceneSize*.5f, 
		    180.0f, 0.0f, NULL);

	/* Do not terrain follow */
	pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_GROUND, 0.0f, NULL);

	/* Do not collide with objects in scene */
	pfuXformerCollision(ViewState->xformer, PFUCOLLIDE_OBJECT, 0.0f, NULL);

	pfGetDCSMat(ViewState->sceneDCS, mat);
	pfuXformerMat(ViewState->xformer, mat);
	break;
    }

    oldMode = mode;
    oldCollMode = collideMode;
}


/******************************************************************************
 *			    Draw Process Routines 
 *************************************************************************** */


/* this will only be initialized in the draw process */
static long Overlay = -1;

void
initPipe(pfPipe *p)
{
    static const pfVec3  clrt = { 0.05f, 0.3f, 0.3f};
    static char msg[] = "Welcome to IRIS Performer";
    
    p; /* suppress compiler warns */

    /* Check for overlay bitplanes */ 
    if (ViewState->input != PFUINPUT_X)
        Overlay = (getgdesc(GD_BITS_OVER_SNG_CMODE) ? OVERDRAW : 
	      	  (getgdesc(GD_BITS_PUP_SNG_CMODE)  ? PUPDRAW  : 0));

    /* Draw introductory message */
    c3f(clrt);
    clear();

#if 0
    /* Message will be draw in in purple - the pfuDrawMessage() color */
    pfuDrawMessage(NULL, msg, PFU_MSG_PIPE, PFU_CENTER_JUSTIFIED, 
		   	0.5f, 0.5f, PFU_FONT_BIG, PFU_RGB);
#endif
    swapbuffers();
}

void
drawUpdate(pfChannel *chan, void *data)
{
    pfPipe *pipe = pfGetChanPipe(chan);
    long pipeId = pfGetId(pipe);

    data; /* suppress compiler warns  */

    /* take snapshot of image */
    if (ViewState->snapImage)
    {
	FILE *fp;
	static  char str[80];
	static  count = 0;
	pfPipe* cpipe = pfGetChanPipe(ViewState->masterChan);
	long    id = pfGetId(cpipe)*10 + count++;

	sprintf(str, "perfly.%d.rgb", id);
        fprintf(stderr, "Saving pipe %d image in file %s\n",
		pfGetId(cpipe), str);

	/* read from FRONT-buffer to get channel stats info */
	readsource(SRC_FRONT);
	pfuSaveImage(str, 0, 0, 
	    ViewState->mouse.winSizeX, 
	    ViewState->mouse.winSizeY);

	/* create info file to go with image */
	sprintf(str, "perfly.%d.info", id);
	if (fp = fopen(str,"w"))
	{
	    float h, v;
	    fprintf(fp, "Viewing parameters for snap %d of perfly\n\n", id);
	    fprintf(fp, "XYZ: %f %f %f\n",
			ViewState->viewCoord.xyz[0],
			ViewState->viewCoord.xyz[1],
			ViewState->viewCoord.xyz[2]);
	    fprintf(fp, "HPR: %f %f %f\n",
			ViewState->viewCoord.hpr[0],
			ViewState->viewCoord.hpr[1],
			ViewState->viewCoord.hpr[2]);
	    fprintf(fp, "NEAR/FAR: %f %f \n",
			ViewState->near, ViewState->far);
	    pfGetChanFOV(ViewState->masterChan, &h, &v);
	    fprintf(fp, "FOV: horiz=%f vert=%f\n", h, v);
	    fclose(fp);
	}

	ViewState->snapImage = 0;
        fprintf(stderr, "Done\n");
    }

    /* 
     * Update new draw modes 
     */
    if (ViewState->drawFlags[pipeId] & UPDATE_MODES)
    {	
	{ /* put this first because it may destroy the old win and
	   * create a new one (in GLX mode) which will require a
	   * window redraw.
	   */
	    static aa=-1;
	    if (aa == -1)
		aa = ViewState->aa;
	    else if (aa != ViewState->aa)
	    {
		if (ViewState->input == PFUINPUT_X)
		{
		    pfPipe *p = pfGetChanPipe(chan);
		    pfuPipeGLXMSConfig(p, ViewState->aa);
		    ViewState->drawFlags[pipeId] |= REDRAW_WINDOW;
		    if (chan == ViewState->masterChan)
			pfuGetGLXWin(p, &(ViewState->xWinInfo));
		}
		if (ViewState->aa)
		    pfAntialias(PFAA_ON);
		else
		    pfAntialias(PFAA_OFF);
		aa = ViewState->aa;
	    }
	}

	if (ViewState->backface)
	    pfCullFace(PFCF_BACK);
	else
	    pfCullFace(PFCF_OFF);

	if (ViewState->wireframe)
	    pfEnable(PFEN_WIREFRAME);
	else
	    pfDisable(PFEN_WIREFRAME);

	if (ViewState->lighting)
	{
	    pfOverride(PFSTATE_ENLIGHTING | PFSTATE_FRONTMTL, 0);
	    pfEnable(PFEN_LIGHTING);
	} 
	else
	{
	    pfDisable(PFEN_LIGHTING);
	    pfOverride(PFSTATE_ENLIGHTING | PFSTATE_FRONTMTL, 1);
	}

	if (ViewState->texture)
	{
	    pfOverride(PFSTATE_ENTEXTURE | PFSTATE_TEXTURE, 0);
	    pfEnable(PFEN_TEXTURE);
	} 
	else
	{
	    pfDisable(PFEN_TEXTURE);
	    pfOverride(PFSTATE_ENTEXTURE | PFSTATE_TEXTURE, 1);
	}
	ViewState->drawFlags[pipeId] &= ~UPDATE_MODES;
    }

    /* Handle REDRAW event */
    if (ViewState->drawFlags[pipeId] & REDRAW_WINDOW)
    {
	/* Only draw message into master channel */
	if (chan == ViewState->masterChan)
	{
	    if (ViewState->redrawOverlay && Overlay)
	    {
		drawOverlayName();
		ViewState->redrawOverlay--;
	    } 
  	    else if (pipeId != pfGetId(pfGetChanPipe(ViewState->masterChan)))
		ViewState->drawFlags[pipeId] &= ~REDRAW_WINDOW;
	} 
	else 
	    ViewState->drawFlags[pipeId] &= ~REDRAW_WINDOW;
    }
}

void
initXformer(void)
{
    ViewState->xformer = pfuNewXformer(pfGetSharedArena());

    pfuXformerAutoInput(ViewState->xformer, ViewState->masterChan, 
	&ViewState->mouse, &ViewState->events);

    xformerMode(ViewState->xformerMode, ViewState->collideMode);
}

void
drawModesChanged(void)
{ 
    long	i;

    for (i=0; i<NumPipes; i++)
	ViewState->drawFlags[i] |= UPDATE_MODES;
}

static void
drawOverlayName(void)
{
#if 0
    static char msg[] = "IRIS Performer";
    static long mapped = 0;
	    
    if (ViewState->input == PFUINPUT_GL)
    {
	drawmode(Overlay);
	if (!mapped) 
	{
	    mapcolor(1, 71, 9, 82);
	    mapcolor(2, 0, 0, 0);
	    gflush();
	    mapped = 1;
	}
    }
    else
    {		
	GLXwinset(ViewState->xWinInfo.dsp, 
		ViewState->xWinInfo.overWin);
		    
	if (mapped != ViewState->xWinInfo.overWin) /* store in X */
	{
	    /* text color - purple */
	    pfuGLXMapcolor(ViewState->xWinInfo.dsp, 
		    ViewState->xWinInfo.overWin, 
		    1, 0.9f, 0.0f, 0.9f);
	    /* shadow color - drk grey */
	    pfuGLXMapcolor(ViewState->xWinInfo.dsp, 
		    ViewState->xWinInfo.overWin, 
		    2, 0.125f, 0.06f, 00.125f);
	    mapped = ViewState->xWinInfo.overWin;
	}
    }
    /* Clear overlay planes in entire window */
    color(0);
    clear();
    pfuDrawMessageCI(ViewState->masterChan, 
	    msg, PFU_MSG_CHAN, PFU_RIGHT_JUSTIFIED, 
	    1.0f, 1.0f, PFU_FONT_MED, 1, 2);
	    
    if (ViewState->input == PFUINPUT_GL)
	drawmode(NORMALDRAW);
    else
	GLXwinset(ViewState->xWinInfo.dsp, 
	    ViewState->xWinInfo.glWin);
#endif
}
