/* polhemus.c - Persia Scheme interface to Polhemus, through SGI's serial port
   Written by: Darrell Anderson 1/96

   When at 19200 baud DIP switches should be 3,5,8 on, all the rest off.
   When at 9600 baud DIP switches should be 1,2,5,8 on, all the rest off.
   See page 13 of the Polhemus FastTrac manual for speed/DIP settings.

   (NOTE: When the Polhemus is connected to IScan's PC, the Polhemus has
   DIP switch 2 on also.)

   See polhemus.h for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include "persia-scm.h"
#include "persia.h"
#include "polhemus.h"
#ifndef USE_STD_SERIAL
#include "/u/becker/src/sgi_vr/Ptvme340/App/hostcmd_sgi.h"
#endif

static short polhemus_numReceivers = POLHEMUS_DEFAULT_NUM_RECEIVERS;
static int polhemus_fd;

static pfDCS *helmet_dcs;
static pfDCS *hand_dcs;

static int polhemus_active = 0;
static int polhemus_initialized = 0;

void polhemus_initialize_scm ();

/*----------------------------------------------------------------
  initialize the polhemus device (called by P_Polhemus_Start)
  --------------------------------------------------------------*/
void polhemus_initialize(char *device, 
			 short baud, 
			 short timeout, 
			 short numReceivers)
{
  struct termio term;
  short i;
  
  /* open the serial port device */
#if USE_STD_SERIAL
  if( (polhemus_fd = open(device, O_RDWR)) < 0 ) {
    fprintf(stderr, "failed to open %d for read/write\n", device);
    exit(1);
  }

  fprintf (stderr, "Opening connection to Polhemus at %d baud\n", (int)baud);
  
  /* configure term */
  
  /* reset io flag */
  term.c_iflag = 0; 
  /* no output post processing */
  term.c_oflag = 0; 
  /* 8 bit, enable receiver */
  term.c_cflag = (CS8 | CREAD | CLOCAL); 
  /* configure for baud rate */
  switch( baud ) {
  case 2400:
    term.c_cflag |= B2400; break;
  case 4800:
    term.c_cflag |= B4800; break;
  case 9600:
    term.c_cflag |= B9600; break;
  case 19200:
    term.c_cflag |= B19200; break;
  default:
    fprintf(stderr, "illegal baud rate: %d\n", baud);
    exit(1);		
  }
  
  /* no local processing */
  term.c_lflag = 0; 
  term.c_line = (unsigned char) 0;
  
  /* no special chars */
  for( i=0 ; i<NCCS ; i++ ) 
    term.c_cc[i] = (unsigned char)0;
  
  /* set character read timeout */
  term.c_cc[VTIME] = (unsigned char)timeout;
  
  /* minimum characters to read is 0 */
  term.c_cc[VMIN] = 0;
  
  /* configure the device */
  if( (ioctl(polhemus_fd, TCSETA, &term)) < 0 ) {
    fprintf(stderr, "failed to set %s characteristics (ioctl)\n", device);
    exit(1);
  }
#else
  hostcmd_init ();
  polhemus_fd = POLHEMUS_DEVICE;
  iusc_init (polhemus_fd, POLHEMUS_BAUD);
#endif
  
  /* other initializtion/setup */
  
  /* put the polhemus in binary mode */
  if( !polhemus_send("f", 1) ) {
    fprintf(stderr, "failed to enable binary mode on polhemus\n");
    exit(1);
  }

  /* set the data receiving mode */
  if( !polhemus_send(PCMD_MODE, strlen(PCMD_MODE)) ) 
  {
    fprintf(stderr, "failed to set mode on polhemus\n");
    exit(1);
  }

#if 1  
  /* set up the specified number of stations */
  polhemus_set_num_receivers( numReceivers );
#endif
}

/*----------------------------------------------------------------
  send numBytes bytes from (*data) to the polhemus
  returns 0 on fail, 1 otherwise
  ----------------------------------------------------------------*/
int polhemus_send( char *data, short numBytes ) {
  short num;

  /* send the data */  
#if USE_STD_SERIAL
  num = write( polhemus_fd, data, numBytes );
#else
  num = write_340iusc( polhemus_fd, data, numBytes );
  if (num != numBytes)
    fprintf (stderr, "Error writing to CommBoard\n");
#endif
  
  /* check if we sent it all */
  if( num != numBytes ) {
    fprintf(stderr, "failed to write to polhemus device (%d of %d reported)\n",
	    num, numBytes);
    return(0);
  }

  /* if we've reached this point something's gone wrong. */
  return(1);
}

/*----------------------------------------------------------------
  read numBytes from the polhemus, storing it in (*data)
  returns 0 on fail, 1 otherwise
  ----------------------------------------------------------------*/
int polhemus_read( char *data, short numBytes ) {
  short num = 0;
  short temp;
  
  /* keep reading until we have enough or something goes wrong */
  while( num < numBytes ) {
    
    /* read what we can */
#if USE_STD_SERIAL
    temp = read( polhemus_fd, data + num, numBytes - num );
#else
    temp = read_340iusc_last_pkt( polhemus_fd, data + num, numBytes - num );
#endif

    /* was there an error? */
    if( temp < 0 ) {
      fprintf(stderr, "failed polhemus read (%d bytes reported)\n",
	      temp);
      return(0);
    }

    /* read EOF before reading 'numBytes' bytes.. 
       (not nec. a timeout!) */
    if( temp == 0 ) {
      printf("polhemus read: EOF after %d of %d, continuing\n", 
	     num, numBytes );
    }

    /* account for the bytes we did read */
    num += temp;
  }
  return(1);
}

/*----------------------------------------------------------------
  change the number of receivers on the fly [*]
  --------------------------------------------------------------*/
void polhemus_set_num_receivers( short numReceivers ) {
  
  /* sanity check */
  if( numReceivers < 1 || numReceivers > 2 ) {
    fprintf(stderr, "number polhemus receivers (%d) invalid\n", numReceivers);
    exit(1);
  }
  
  /* remember how many stations we're dealing with */
  polhemus_numReceivers = numReceivers;
  
  /* only do the setup if the polhemus has been previously initialized */
  /* (otherwise it will occur on initialization) */
  if( polhemus_initialized ) {
    /* see "l" on page 129 in 3-SPACE user manual (station on/off states) */
    
    /* station 1 is always on (head) */
    if( !polhemus_send(PCMD_ENABLE_1, strlen(PCMD_ENABLE_1)) )
      fprintf(stderr, "failed to enable polhemus station 1\n");
    
    /* enable or disable station 2, depending on numReceivers */
    if( numReceivers >= 2 ) {
      if( !polhemus_send(PCMD_ENABLE_2, strlen(PCMD_ENABLE_2)) )
	fprintf(stderr, "failed to enable polhemus station 2\n");
    } else {
      if( !polhemus_send(PCMD_DISABLE_2, strlen(PCMD_DISABLE_2)) )
	fprintf(stderr, "failed to disable polhemus station 2\n");
    }
  }
}

/*----------------------------------------------------------------
  update global head, hand matrices, called as a pre-frame hook [*]
  --------------------------------------------------------------*/
void polhemus_c_hook(void) {
  polhemus_record data[2];
  pfMatrix polhemus_in_xmit_coords;
  int i, j;
 
  /* request record from the polhemus (skip error checking) */
  polhemus_send( PCMD_REQUEST_RECORD, 1 );
  
  /* read the forthcoming record (byte order is X86 endian) 
     note: the +1/-1 deal accounts for an extra byte at the 
     start of the structure */
  polhemus_read(((char*)(&(data[0])))+1, 
		sizeof(polhemus_record)-1 );
  
  /* if we have two (or more!) receivers, read a second record */
  if( polhemus_numReceivers >= 2 )
    polhemus_read(((char*)(&(data[1])))+1, 
		  sizeof(polhemus_record)-1 );
  
  /* deal with all of the data records */
  for( i=0 ; i<polhemus_numReceivers ; i++ )
    {

      /* When testing this, by setting the headhand.scm matrices to the 
	 identity matrix, translation in all directions seems fine,
	 but rotation of the x-axis is backwards and rotation of the
	 z-axis is backwards.  -mccallum. */

      /* swap the endian-ness of the data portion */
      for( j=4 ; j<sizeof(polhemus_record) ; j+= 4 )
	reverse_byte_order( ((unsigned char *)data) + j );
    
      /* interpret into a rotation/translation matrix representing
	 sensor location/orientation with respect to transmitter */
#if (POLHEMUS_MODE == QUATERNION_MODE)
      /* NOTE!  This is the transpose of the matrix given on page 163 of 
	 the Polhemus FastTrac manual.  While the formula on page 163 
	 gives the X, Y, and Z directional cosines in columns, it looks 
	 like the SGI expects the X, Y, and Z directional cosines in 
	 rows. */
      polhemus_in_xmit_coords[0][0] = data[i].q0 * data[i].q0 + data[i].q1 * 
	data[i].q1 - data[i].q2 * data[i].q2 - data[i].q3 * data[i].q3;
      polhemus_in_xmit_coords[0][1] =  2 * 
	(data[i].q3 * data[i].q0 + data[i].q1 * data[i].q2);
      polhemus_in_xmit_coords[0][2] = 2 * 
	(data[i].q1 * data[i].q3 - data[i].q0 * data[i].q2);
    
      polhemus_in_xmit_coords[1][0] = 2 * 
	(data[i].q1 * data[i].q2 - data[i].q0 * data[i].q3);
      polhemus_in_xmit_coords[1][1] = data[i].q0 * data[i].q0 - data[i].q1 * 
	data[i].q1 + data[i].q2 * data[i].q2 - data[i].q3 * data[i].q3;;
      polhemus_in_xmit_coords[1][2] = 2 * 
	(data[i].q1 * data[i].q0 + data[i].q3 * data[i].q2);
    
      polhemus_in_xmit_coords[2][0] = 2 * 
	(data[i].q1 * data[i].q3 + data[i].q0 * data[i].q2);
      polhemus_in_xmit_coords[2][1] = 2 * 
	(data[i].q2 * data[i].q3 - data[i].q0 * data[i].q1);
      polhemus_in_xmit_coords[2][2] = data[i].q0 * data[i].q0 - data[i].q1 * 
	data[i].q1 - data[i].q2 * data[i].q2 + data[i].q3 * data[i].q3;
#elif (POLHEMUS_MODE == EULER_MODE)
      /* expects heading (azimuth), pitch (elevation), roll */
      /* however azim, roll, elev seems to work */
      pfMakeEulerMat (polhemus_in_xmit_coords, 
		      data[i].azim,
		      data[i].elev,
		      -data[i].roll);
      /* This doesn't actually work because the SGI uses a different 
	 Euler system. */
      /* printf ("azim %6.1f elev %6.1f roll %6.1f\n",
	 data[i].azim, data[i].elev, data[i].roll); */
#elif (POLHEMUS_MODE == DIR_COS_MODE)
      polhemus_in_xmit_coords[0][0] = data[i].x1;
      polhemus_in_xmit_coords[1][0] = data[i].x2;
      polhemus_in_xmit_coords[2][0] = data[i].x3;
      
      polhemus_in_xmit_coords[0][1] = data[i].y1;
      polhemus_in_xmit_coords[1][1] = data[i].y2;
      polhemus_in_xmit_coords[2][1] = data[i].y3;
      
      polhemus_in_xmit_coords[0][2] = data[i].z1;
      polhemus_in_xmit_coords[1][2] = data[i].z2;
      polhemus_in_xmit_coords[2][2] = data[i].z3;
#else
#error bad POLHEMUS_MODE
#endif /* POLHEMUS_MODE == x */

      /* Fill in translation. (data is in cm, convert to meters) */
      polhemus_in_xmit_coords[3][0] = data[i].x / 100.0;
      polhemus_in_xmit_coords[3][1] = data[i].y / 100.0;
      polhemus_in_xmit_coords[3][2] = data[i].z / 100.0;  
      polhemus_in_xmit_coords[0][3] = 0;
      polhemus_in_xmit_coords[1][3] = 0;
      polhemus_in_xmit_coords[2][3] = 0;
      polhemus_in_xmit_coords[3][3] = 1;
    
      switch( data[i].station )
	{
	case '0':
	  /* hand */
	  pfDCSMat(hand_dcs, polhemus_in_xmit_coords);
	  break;
	case '1':
	  /* helmet */
	  pfDCSMat(helmet_dcs, polhemus_in_xmit_coords);
	  break;
	default:
	  fprintf(stderr,"invalid station number read from polhemus (%d)\n",
		  (int)data[i].station);
	  break;
	}
    }
}

/*----------------------------------------------------------------
  master polhemus initialization (and activation) funtion, 
  called from scheme [*]
  ----------------------------------------------------------------*/
Object P_Polhemus_Start(void) {
  if( !ViewState ) 
    Primitive_Error("ps:init-cameras not yet called");
  
  polhemus_initialize_scm ();
  /* do this each time the Polhemus is re-enabled */
  if( !polhemus_active ) {
    polhemus_active = 1;
    persia_add_pre_frame_hook(polhemus_c_hook);
  }
  
  /* do this only once! */
  if( !polhemus_initialized ) {
    polhemus_initialized = 1;
    polhemus_initialize(POLHEMUS_DEVICE, POLHEMUS_BAUD, POLHEMUS_TIMEOUT, 
			polhemus_numReceivers );
  }
  return Void;
}

/*----------------------------------------------------------------
  master polhemus kill(suspend really) function, called from scheme [*]
  ----------------------------------------------------------------*/
Object P_Polhemus_Stop(void) {
  if( !ViewState ) 
    Primitive_Error("ps:init-cameras not yet called");
  /* disable the polhemus */
  if( polhemus_active ) {
    polhemus_active = 0;
    persia_remove_pre_frame_hook(polhemus_c_hook);
  }
  return Void;
}

/*----------------------------------------------------------------
  change the number of receivers on the fly [*]
  ----------------------------------------------------------------*/
Object P_Polhemus_SetNumReceivers(Object num) {
  Check_Type(num, T_Fixnum);
  if( FIXNUM(num) < 1 || FIXNUM(num) > 2 )
    Primitive_Error("number of polhemus receivers is out of range");
  polhemus_set_num_receivers( FIXNUM(num) );
  return Void;
}

/*----------------------------------------------------------------
  make above scheme functions available (a la Define_Primitive) [*]
  ----------------------------------------------------------------*/
void polhemus_initialize_scm ()
{
  static int done = 0;
  if (!done)
    {
      done = 1;
      /* create the global helmet and hand structures */
      helmet_dcs = pfNewDCS();
      hand_dcs = pfNewDCS();
      /* make them available in scheme */
      Define_Pfo((pfObject*)helmet_dcs, "ps:*head-dcs*");
      Define_Pfo((pfObject*)hand_dcs, "ps:*hand-dcs*");
    }
}

void polhemus_init(void)
{
  /* make the functions available in scheme */
  Define_Primitive(P_Polhemus_Start, "ps:headhand-start", 0, 0, EVAL);
  Define_Primitive(P_Polhemus_Stop, "ps:headhand-stop", 0, 0, EVAL);
  Define_Primitive(P_Polhemus_SetNumReceivers, 
		   "ps:headhand-set-num-receivers", 1, 1, EVAL);
}

void init_polhemus ()
{
  polhemus_init ();
}

void reverse_byte_order( unsigned char *word ) {
  unsigned char temp;

  temp = word[0];
  word[0] = word[3];
  word[3] = temp;
  temp = word[1];
  word[1] = word[2];
  word[2] = temp;
}


