/*
polhemus_sgi.h
Darrell Anderson 1/96

Polhemus communication code.  Assumes sgi is connected to the polhemus
via POLHEMUS_DEVICE (/dev/[rs232] defined below).  Allows multiple
simultaneous receivers.

The interface is modelled after that presented in flock.c, in order to
be (somewhat) interchangable. */

#ifndef __POLHEMUS_SGI_H__
#define __POLHEMUS_SGI_H__

/* 0 means use the standard SGI serial port.
   1 means use the "CommBoard" with Tim Becker's code. */
#define USE_STD_SERIAL 0

#if (USE_STD_SERIAL == 1)
/* This is for the standard SGI serial port */
#define POLHEMUS_DEVICE "/dev/ttyd2"
/* NOTE:  If you change the speed you must change the DIP switches on the
   back of the Polhemus.  See page 13 of the Polhemus FastTrac manual. */
/* #define POLHEMUS_BAUD 19200 
   This is too fast.  We get noise when reading from both polhemus and kart */
#define POLHEMUS_BAUD 2400
#else
/* This is for the CommBoard with Tim Becker's code. */
#define POLHEMUS_DEVICE 0
#define POLHEMUS_BAUD 57600
#endif

#define POLHEMUS_TIMEOUT 1
#define POLHEMUS_DEFAULT_NUM_RECEIVERS 1


/* polhemus commands */
#define QUATERNION_MODE 1
#define EULER_MODE 2
#define DIR_COS_MODE 3

#define POLHEMUS_MODE QUATERNION_MODE
/* Setting this to DIR_COS_MODE gives errors 
   `invalid station number read from polhemus (44)'
   -mccallum 6/13/96 */

#if (POLHEMUS_MODE == 1)
#define PCMD_MODE "O1,2,11\rO2,2,11\r" 
#elif (POLHEMUS_MODE == 2)
#define PCMD_MODE "O1,2,4\rO2,2,4\r"
#elif (POLHEMUS_MODE == 3)
#define PCMD_MODE "O1,2,5,6,7\rO2,2,5,6,7"
#else
#error Bad POLHEMUS_MODE
#endif

/* see page 98 of the 3-SPACE manual for specifics, this sets the  
   /* first two stations to report data in x,y,z,quaternion format */
#define PCMD_REQUEST_RECORD "P"
#define PCMD_ENABLE_1 "l1,1\r" /* enable/disable stations */
#define PCMD_ENABLE_2 "l2,1\r"
#define PCMD_DISABLE_2 "l2,0\r"

/* #define PCMD_METRIC "u" */  /* NYI */
/* #define PCMD_ENGLISH "U" */ /* NYI */

/*----------------------------------------------------------------
initialize the polhemus device (called by P_Polhemus_Start)
----------------------------------------------------------------*/
void polhemus_initialize(char *device, short baud, short timeout, 
			 short numReceivers);

/*----------------------------------------------------------------
send numBytes bytes from (*data) to the polhemus
returns 0 on fail, 1 otherwise
----------------------------------------------------------------*/
int polhemus_send( char *data, short numBytes );

/*----------------------------------------------------------------
read numBytes from the polhemus, storing it in (*data)
returns 0 on fail, 1 otherwise
----------------------------------------------------------------*/
int polhemus_read( char *data, short numBytes );

/*----------------------------------------------------------------
change the number of receivers on the fly
----------------------------------------------------------------*/
void polhemus_set_num_receivers( short numReceivers );

/*----------------------------------------------------------------
update global head and hand matrices, called as a pre-frame hook 
P_Polhemus_Start must have been previously called for this to work.
----------------------------------------------------------------*/
void polhemus_c_hook(void);

/*----------------------------------------------------------------
master polhemus initialization (and activation) funtion, called from  
scheme Polhemus_Init must have been previously called for this to work.
----------------------------------------------------------------*/
Object P_Polhemus_Start(void);

/*----------------------------------------------------------------
master polhemus kill(suspend really) function, called from scheme 
P_Polhemus_Start must have been previously called for this to work.
----------------------------------------------------------------*/
Object P_Polhemus_Stop(void);

/*----------------------------------------------------------------
change the number of receivers on the fly, called from scheme 
P_Polhemus_Init must have been previously called for this to work.
----------------------------------------------------------------*/
Object P_Polhemus_SetNumReceivers(Object num);

/*----------------------------------------------------------------
make the global hand/head variables and above P_ functions available 
----------------------------------------------------------------*/
void Polhemus_Init(void);

/*--------------
reverse the byte order in a word X86->'normal' (:
--------------*/
void reverse_byte_order( unsigned char *word );

/* the data format we request from the polhemus: */
typedef struct polhemus_record {
	/* unused byte.. needed to keep rest of struct continuous and */
	/* aligned (otherwise C will skip a byte to align the float) */
	unsigned char unused;

	/* record header */
	unsigned char record_type;
	unsigned char station;
	unsigned char sys_err;
	
	/* position */
	float x,y,z;

#if (POLHEMUS_MODE == QUATERNION_MODE)
	/* quaternion */
	float q0,q1,q2,q3; 
#elif (POLHEMUS_MODE == EULER_MODE)
	float azim, elev, roll;
#else
	/* direction cosines */
	float x1, x2, x3, y1, y2, y3, z1, z2, z3;
#endif

} polhemus_record;

#endif /* __POLHEMUS_SGI_H__ */




